# -*- coding: utf-8 -*-
import numpy as np
from numba import jit

from ..sparse.csr import CSR


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def signed_topo_CSR(conn: CSR) -> CSR:
    return None


@jit(nopython=True, nogil=True)
def signed_topo_flattened(a: np.ndarray, path: np.ndarray):
    a = np.abs(a)
    a_max = np.max(a)
    a_min = np.min(a)
    res = a + 1 - a_min
    found = np.full(a_max + 1 - a_min, 0, dtype=a.dtype)
    for i in range(len(res) - 1, -1, -1):
        if found[a[path[i]] - a_min] == 0:
            found[a[path[i]] - a_min] = 1
            res[path[i]] *= -1
    return res


@jit(nopython=True, nogil=True, cache=True)
def signed_topo_bulk(a: np.ndarray, path: np.ndarray):
    a = np.abs(a)
    a_max = np.max(a)
    a_min = np.min(a)
    res = a + 1 - a_min
    I, J = a.shape
    found = np.full(a_max + 1 - a_min, 0, dtype=a.dtype)
    for i in range(I - 1, -1, -1):
        for j in range(J):
            if found[a[path[i], j] - a_min] == 0:
                found[a[path[i], j] - a_min] = 1
                res[path[i], j] *= -1
    return res


@jit(nopython=True, nogil=True, cache=True)
def unsign_topo(signed: np.ndarray):
    return np.abs(signed) - 1


@jit(nopython=True, nogil=True, cache=True)
def topo_to_adj(topo: np.ndarray):
    nN = np.max(topo) + 1
    res = np.zeros((nN, nN), dtype=topo.dtype)
    nE, nNE = topo.shape
    for iE in range(nE):
        for iNE in range(nNE):
            for jNE in range(nNE):
                res[topo[iE, iNE], topo[iE, jNE]] = 1
                res[topo[iE, jNE], topo[iE, iNE]] = 1
    return res



from typing import Union, Tuple

import numpy as np
from numpy.linalg import LinAlgError
from numba import njit

__cache = True


__all__ = ["linsolve", "reduce", "backsub", "npsolve"]


def linsolve(
    A: np.ndarray,
    B: np.ndarray,
    *,
    presc_bool: np.ndarray = None,
    presc_val: np.ndarray = None,
    method: str = "numpy",
    inplace: bool = False
) -> Union[np.ndarray, Tuple[np.ndarray]]:
    """
    Solves a linear system of equations using various methods. It supports prescription of values.

    Parameters
    ----------
    A : numpy.ndarray
        The coefficient matrix.
    B : numpy.ndarray
        One more right hand sides.
    presc_bool : numpy.ndarray, Optional
        An 1d NumPy array of booleans to indicate which unknowns are prescribed.
        Default is None.
    presc_val : numpy.ndarray, Optional
        An 1d NumPy array of floats of prescribed values. It should be used together
        with 'presc_bool'. Default is None.
    method : str, Optional
        The method to use. Valid options are:
        * 'numpy' :  uses NumPy's solver for dense scenarios. In this case prescribed
           values are not available. This is the same as to call `numpy.linalg.solve` directly.
        * 'Jordan' : Jordan elimination
        * 'Gauss-Jordan' : Gauss-Jordan elimination
    inplace : bool, Optional
        To allow for in-place modifications of the inputs 'A' and 'B' not. If False, a
        copy is created. Default is False.

    Note
    ----
    Methods 'Jordan' and 'Gauss-Jordan' use elimination to account for constraints on the unknown
    variables, hence yield exact results. These algorithms doesn't scale too well and are not suggested
    in a production enviroment, but they are handy to measure the error of a penalty function approach
    or something else. However, their implementation relies on JIT-compiled code using Numba, and
    for small and medium sized problems, they should do the job in a reasonable amount of time.

    Returns
    -------
    numpy.ndarray
        The solution with a shape matching input 'B'.
    numpy.ndarray, Optional
        The residual vector, if the solution is constrained.

    Example
    -------
    To solve a simple system of equation:

    >>> from sigmaepsilon.math.linalg.solve import linsolve
    >>> A = np.array([[3, 1, 2], [1, 1, 1], [2, 1, 2]], dtype=float)
    >>> B = np.array([11, 6, 10], dtype=float)
    >>> X = linsolve(A, B, method='Gauss-Jordan')
    >>> X.flatten()
    array([1., 2., 3.])

    Of course, in this case you could have used NumPy's solver, but if you want to
    have a few unknowns prescribed, you can make use of the direct solvers presented
    here.

    >>> pb = np.array([False, True, False], dtype=bool)
    >>> pv = np.array([0, 3, 0], dtype=float)
    >>> X, R = linsolve(A, B, method='Gauss-Jordan', presc_bool=pb, presc_val=pv)

    In this case, the first argument is the constrained solution vector

    >>> X.flatten()
    array([1. , 3. , 2.5])

    and the second one is the vector of resudials due to the constraints

    >>> R.flatten()
    array([0. , 0.5, 0. ])

    the meaning of which becomes clear if you compare the image of the constrained
    solution you've just calculated against the input B

    >>> np.ravel(A @ X)  # ravel is used to flatten the column array
    array([11. ,  6.5, 10. ])

    >>> B
    array([11.,  6., 10.])
    """
    # FIXME : check if presc_bool and presc_val have matching shape to A and B
    # FIXME : check if presc_bool is a boolean array
    if method == "numpy":
        assert (
            presc_bool is None
        ), "Method '{}' does not support prescribed values.".format(method)
        return npsolve(A, B)
    elif method in ["Jordan", "Gauss-Jordan"]:
        fnc = _Jordan if method == "Jordan" else _GaussJordan
        try:
            nEQ = len(B)
            if len(B.shape) == 1:
                B = B.reshape((nEQ, 1))
            pre = presc_val is not None
            if not pre:
                presc_bool = np.zeros((nEQ,), dtype=bool)
                presc_val = np.zeros((nEQ,), dtype=np.float64)
            if inplace:
                res = backsub(*fnc(A, B, presc_bool, presc_val), presc_bool, presc_val)
            else:
                res = backsub(
                    *fnc(A.copy(), B.copy(), presc_bool, presc_val),
                    presc_bool,
                    presc_val
                )
            if pre:
                return res
            else:
                return res[0]
        except Exception:
            raise LinAlgError("The matrix is singular!")
    else:
        raise AttributeError("Invalid method name '{}'.".format(method))


def reduce(
    A: np.ndarray,
    B: np.ndarray,
    presc_bool: np.ndarray = None,
    presc_val: np.ndarray = None,
    method="Gauss-Jordan",
    inplace=False,
):
    fnc = None
    if method == "Gauss-Jordan":
        fnc = _GaussJordan
    elif method == "Jordan":
        fnc = _Jordan
    if fnc is not None:
        try:
            if presc_bool is None:
                nEQ = len(B)
                presc_bool = np.zeros((nEQ,), dtype=np.int64)
                presc_val = np.zeros((nEQ,), dtype=np.float64)
            if inplace:
                return fnc(A, B, presc_bool, presc_val)
            else:
                return fnc(A.copy(), B.copy(), presc_bool, presc_val)
        except Exception:
            raise LinAlgError("The matrix is singular!")
    else:
        raise AttributeError("Invalid method name '{}'.".format(method))


@njit(nogil=True, cache=__cache)
def npsolve(A, b):
    return np.linalg.solve(A, b)


@njit(nogil=True, cache=__cache)
def _Jordan(
    A: np.ndarray, B: np.ndarray, presc_bool: np.ndarray, presc_val: np.ndarray
):
    nEQ = len(B)
    for iEQ in range(nEQ):
        if presc_bool[iEQ] == True:
            for iROW in range(iEQ, nEQ):
                B[iROW] -= A[iROW, iEQ] * presc_val[iEQ]
                A[iROW, iEQ] = 0.0
        else:
            pivot = A[iEQ, iEQ]
            if abs(pivot) < 1e-12:
                raise Exception("The matrix is singular!")
            for iROW in range(nEQ):
                factor = A[iROW, iEQ] / pivot
                if iROW == iEQ or abs(factor) < 1e-12:
                    continue
                for iCOL in range(nEQ):
                    A[iROW, iCOL] -= factor * A[iEQ, iCOL]
                B[iROW] -= factor * B[iEQ]
    return A, B


@njit(nogil=True, cache=__cache)
def _GaussJordan(
    A: np.ndarray, B: np.ndarray, presc_bool: np.ndarray, presc_val: np.ndarray
):
    nEQ = len(B)
    for iEQ in range(nEQ):
        if presc_bool[iEQ] == True:
            for iROW in range(iEQ, nEQ):
                B[iROW] -= A[iROW, iEQ] * presc_val[iEQ]
                A[iROW, iEQ] = 0.0
        else:
            if iEQ < nEQ - 1:
                pivot = A[iEQ, iEQ]
                if abs(pivot) < 1e-12:
                    raise Exception("The matrix is singular!")
                for iROW in range(iEQ + 1, nEQ):
                    factor = A[iROW, iEQ] / pivot
                    if abs(factor) < 1e-12:
                        continue
                    for iCOL in range(nEQ):
                        A[iROW, iCOL] -= factor * A[iEQ, iCOL]
                    B[iROW] -= factor * B[iEQ]
    return A, B


@njit(nogil=True, cache=__cache)
def backsub(
    A: np.ndarray, B: np.ndarray, presc_bool: np.ndarray, presc_val: np.ndarray
):
    nEQ, nRHS = B.shape
    X = np.zeros((nEQ, nRHS), dtype=np.float64)
    R = np.zeros((nEQ, nRHS), dtype=np.float64)
    resid = np.zeros(nRHS, dtype=np.float64)
    for iEQ in range(nEQ - 1, -1, -1):
        pivot = A[iEQ, iEQ]
        resid[:] = B[iEQ]
        if iEQ < nEQ - 1:
            for iCOL in range(iEQ + 1, nEQ):
                resid -= A[iEQ, iCOL] * X[iCOL]
        if presc_bool[iEQ] == True:
            X[iEQ] = presc_val[iEQ]
            R[iEQ] = -resid
        else:
            X[iEQ] = resid / pivot
    return X, R



from enum import Enum
import operator as op
from typing import TypeVar, Callable

from sigmaepsilon.core.kwargtools import getasany

from .function import Function


__all__ = ["Equality", "InEquality", "Relation"]


class Relations(Enum):
    eq = "="
    gt = ">"
    ge = ">="
    lt = "<"
    le = "<="

    def to_op(self):
        return _rel_to_op[self]


_rel_to_op = {
    Relations.eq: op.eq,
    Relations.gt: op.gt,
    Relations.ge: op.ge,
    Relations.lt: op.lt,
    Relations.le: op.le,
}

RelationType = TypeVar("RelationType", str, Relations, Callable)


class Relation(Function):
    """
    Base class for relations.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.op = None
        self.opfunc = None
        op = getasany(["op", "operator"], None, **kwargs)
        if op:
            if isinstance(op, str):
                self.op = Relations(op)
            elif isinstance(op, Relations):
                self.op = op
            elif isinstance(op, Callable):
                self.opfunc = op
                self.op = None
        else:
            self.op = Relations.eq
        if op and isinstance(self.op, Relations):
            self.opfunc = self.op.to_op()
        self.slack = 0

    @property
    def operator(self) -> Callable:
        """
        Returns the associated operator.
        """
        return self.op

    def to_eq(self):
        raise NotImplementedError

    def relate(self, *args, **kwargs):
        """
        Relates an input and returns True if it is feasible.
        """
        return self.opfunc(self.f0(*args, **kwargs), 0)


class Equality(Relation):
    """
    Class for equality constraints.

    Example
    -------
    >>> import sympy as sy
    >>> variables = ['x1', 'x2', 'x3', 'x4']
    >>> x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    >>> eq1 = Equality(x1 + 2*x3 + x4 - 4, variables=syms)
    >>> eq2 = Equality(x2 + x3 - x4 - 2, variables=syms)
    """

    def __init__(self, *args, **kwargs):
        kwargs["op"] = Relations.eq
        super().__init__(*args, **kwargs)

    def to_eq(self):
        return self


class InEquality(Relation):
    """
    Class for inequality constraints.

    Examples
    --------
    >>> gt = InEquality('x + y', op='>')
    >>> gt([0.0, 0.0])
    0.0

    >>> ge = InEquality('x + y', op='>=')
    >>> ge([0.0, 0.0])
    0.0

    >>> le = InEquality('x + y', op=lambda x, y: x <= y)
    >>> le([0.0, 0.0])
    0.0

    >>> lt = InEquality('x + y', op=lambda x, y: x < y)
    >>> lt([0.0, 0.0])
    0.0
    """

    def __init__(self, *args, **kwargs):
        op = getasany(["op", "operator"], None, **kwargs)
        if not op:
            raise ValueError("An operator must be defined.")
        super().__init__(*args, **kwargs)

    def to_eq(self):
        raise NotImplementedError



from numpy import ndarray


__all__ = ["LSApproximator"]


class LSApproximator:
    """
    A class to perform least-squares approximations on known data.

    Given :math:`N` points located at :math:`\mathbf{x}_i` in :math:`\mathbb{R}^d`
    where :math:`i \in [1 \dots N]`. The returned fit function approximates the given
    values :math:`f_i` at :math:`\mathbf{x}_i` in the least-squares sence with the
    error functional

    .. math::
        \sum_{i} \\biggr[ || f \left( \mathbf{x}_i \\right) - f_i || \\biggr] ^2

    where :math:`f` is taken from :math:`\Pi_{m}^d`, the space of polynomials of
    total degree :math:`m` in :math:`d` spatial dimensions.

    Note
    ----
    The fit function can have an approximation or regression behaviour,
    depending on the dataset and the degree of the polynomial.
    """

    def fit(self, coords: ndarray, values: ndarray, deg: int, order: int) -> None:
        """
        Creates the fitting function.

        Parameters
        ----------
        coords: numpy.ndarray
            The coordinates of the points where data is known as a 2d NumPy array.
        values: numpy.ndarray
            The known data.
        """
        pass

    def approximate(self, coords: ndarray) -> ndarray:
        """
        Approcimates the field variables onto the desired coordinates.

        Parameters
        ----------
        coords: numpy.ndarray
            The coordinates of the target points of approximation.
        """
        pass



import numpy as np
from collections import namedtuple


__all__ = [
    "Quadrature",
    "gauss_points",
    "gauss_points_1d",
    "gauss_points_2d",
    "gauss_points_3d",
]


Quadrature = namedtuple("QuadratureRule", ["inds", "pos", "weight"])


def gauss_points(*args):
    """
    Returns data for numerical integration on an N-dimensional unit square using
    the Gauss-Legendre rule for 1d, 2d and 3d scenarios. The implementation using
    `numpy.polynomial.legendre.leggauss` repeatedly for higher dimensions.

    Parameters
    ----------
    *args : tuple
        An integer for each dimension (see the examples).

    Returns
    -------
    numpy.ndarray
        A NumPy array. For one dimension, this output includes the locations and
        the weights as well, for higher dimensions this only contains locations,
        and the weights are returned as a separate array.
    numpy.ndarray
        The weights as a separate output. Only for dimensions greater than 1.

    Examples
    --------
    For a one dimensional case, there is only one output. The first row of the returned array
    are the locations, the second one are the weights.

    >>> from sigmaepsilon.math.numint import gauss_points
    >>> gauss_points(2)
    array([[-0.57735027,  0.57735027],
           [ 1.        ,  1.        ]])

    For 2d:

    >>> gauss_points(2, 2)
    (array([[-0.57735027, -0.57735027],
           [-0.57735027,  0.57735027],
           [ 0.57735027, -0.57735027],
           [ 0.57735027,  0.57735027]]), array([1., 1., 1., 1.]))

    For 3d:

    >>> gauss_points(2, 2, 2)
    (array([[-0.57735027, -0.57735027, -0.57735027],
           [-0.57735027, -0.57735027,  0.57735027],
           [-0.57735027,  0.57735027, -0.57735027],
           [-0.57735027,  0.57735027,  0.57735027],
           [ 0.57735027, -0.57735027, -0.57735027],
           [ 0.57735027, -0.57735027,  0.57735027],
           [ 0.57735027,  0.57735027, -0.57735027],
           [ 0.57735027,  0.57735027,  0.57735027]]), array([1., 1., 1., 1., 1., 1., 1., 1.]))
    """
    nD = len(args)
    if nD == 1:
        return gauss_points_1d(args[0])
    elif nD == 2:
        return gauss_points_2d(args)
    elif nD == 3:
        return gauss_points_3d(args)
    else:
        raise NotImplementedError(
            "Currently there is no implementation for dimensions higher than 3."
        )


def gauss_points_1d(NumPoints):
    x, w = np.polynomial.legendre.leggauss(NumPoints)
    v = np.zeros([2, NumPoints])
    v[0, :] = x
    v[1, :] = w
    return v


def gauss_points_2d(NumPoints):
    nGaus = NumPoints[0] * NumPoints[1]
    QuadraturePos = np.zeros((nGaus, 2))
    QuadratureWeight = np.zeros((nGaus))
    quad1 = gauss_points_1d(NumPoints[0])
    quad2 = gauss_points_1d(NumPoints[1])
    g = 0
    for gi in np.nditer(quad1, flags=["external_loop"], order="F"):
        for gj in np.nditer(quad2, flags=["external_loop"], order="F"):
            QuadraturePos[g] = np.array([gi[0], gj[0]])
            QuadratureWeight[g] = gi[1] * gj[1]
            g += 1
    return QuadraturePos, QuadratureWeight


def gauss_points_3d(NumPoints):
    nGaus = NumPoints[0] * NumPoints[1] * NumPoints[2]
    QuadraturePos = np.zeros((nGaus, 3))
    QuadratureWeight = np.zeros((nGaus))
    quad1 = gauss_points_1d(NumPoints[0])
    quad2 = gauss_points_1d(NumPoints[1])
    quad3 = gauss_points_1d(NumPoints[2])
    g = 0
    for gi in np.nditer(quad1, flags=["external_loop"], order="F"):
        for gj in np.nditer(quad2, flags=["external_loop"], order="F"):
            for gk in np.nditer(quad3, flags=["external_loop"], order="F"):
                QuadraturePos[g] = np.array([gi[0], gj[0], gk[0]])
                QuadratureWeight[g] = gi[1] * gj[1] * gk[1]
                g += 1
    return QuadraturePos, QuadratureWeight



from typing import Tuple

from numpy import ndarray
from numba import njit


def _mls_approx(*args, dim: int = None, **kwargs):
    if dim == 1:
        return _mls_approx_1d(*args, **kwargs)
    elif dim == 2:
        return _mls_approx_2d(*args, **kwargs)
    elif dim == 3:
        return _mls_approx_3d(*args, **kwargs)
    else: # pragma: no cover
        raise NotImplementedError(
            "Only 1, 2 and 3 dimensional pointclouds are supported."
        )


def _mls_approx_1d(
    invA: ndarray,
    B: ndarray,
    b: ndarray,
    V: ndarray,
    Adx: ndarray,
    Adxx: ndarray,
    Bdx: ndarray,
    Bdxx: ndarray,
    bdx: ndarray,
    bdxx: ndarray,
    g: bool = True,
    H: bool = True,
) -> Tuple[ndarray, ndarray, ndarray]:
    gamma = invA @ b
    f = _mls_f(gamma, B, V)

    fdx, fdxx = (None,) * 2

    if g:
        gammadx = invA @ (bdx - Adx @ gamma)
        fdx = _mls_g_1d(gamma, gammadx, B, Bdx, V)

    if H:
        fdxx = _mls_H_1d(
            invA,
            bdxx,
            Adx,
            Adxx,
            gamma,
            gammadx,
            B,
            Bdx,
            Bdxx,
            V,
        )

    return f, fdx, fdxx


def _mls_approx_2d(
    invA: ndarray,
    B: ndarray,
    b: ndarray,
    V: ndarray,
    Adx: ndarray,
    Ady: ndarray,
    Adxx: ndarray,
    Adyy: ndarray,
    Adxy: ndarray,
    Bdx: ndarray,
    Bdy: ndarray,
    Bdxx: ndarray,
    Bdyy: ndarray,
    Bdxy: ndarray,
    bdx: ndarray,
    bdy: ndarray,
    bdxx: ndarray,
    bdyy: ndarray,
    bdxy: ndarray,
    g: bool = True,
    H: bool = True,
) -> Tuple[ndarray, ndarray, ndarray, ndarray, ndarray, ndarray]:
    gamma = invA @ b
    f = _mls_f(gamma, B, V)

    fdx, fdy, fdxx, fdyy, fdxy = (None,) * 5

    if g:
        gammadx = invA @ (bdx - Adx @ gamma)
        gammady = invA @ (bdy - Ady @ gamma)
        fdx, fdy = _mls_g_2d(gamma, gammadx, gammady, B, Bdx, Bdy, V)

    if H:
        fdxx, fdyy, fdxy = _mls_H_2d(
            invA,
            bdxx,
            bdyy,
            bdxy,
            Adx,
            Ady,
            Adxx,
            Adyy,
            Adxy,
            gamma,
            gammadx,
            gammady,
            B,
            Bdx,
            Bdy,
            Bdxx,
            Bdyy,
            Bdxy,
            V,
        )

    return f, fdx, fdy, fdxx, fdyy, fdxy


def _mls_approx_3d(
    invA: ndarray,
    B: ndarray,
    b: ndarray,
    V: ndarray,
    Adx: ndarray,
    Ady: ndarray,
    Adz: ndarray,
    Adxx: ndarray,
    Adyy: ndarray,
    Adzz: ndarray,
    Adxy: ndarray,
    Adxz: ndarray,
    Adyz: ndarray,
    Bdx: ndarray,
    Bdy: ndarray,
    Bdz: ndarray,
    Bdxx: ndarray,
    Bdyy: ndarray,
    Bdzz: ndarray,
    Bdxy: ndarray,
    Bdxz: ndarray,
    Bdyz: ndarray,
    bdx: ndarray,
    bdy: ndarray,
    bdz: ndarray,
    bdxx: ndarray,
    bdyy: ndarray,
    bdzz: ndarray,
    bdxy: ndarray,
    bdxz: ndarray,
    bdyz: ndarray,
    g: bool = True,
    H: bool = True,
) -> Tuple[
    ndarray,
    ndarray,
    ndarray,
    ndarray,
    ndarray,
    ndarray,
    ndarray,
    ndarray,
    ndarray,
    ndarray,
]:
    gamma = invA @ b
    f = _mls_f(gamma, B, V)

    fdx, fdy, fdz, fdxx, fdyy, fdzz, fdxy, fdxz, fdyz = (None,) * 9

    if g:
        gammadx = invA @ (bdx - Adx @ gamma)
        gammady = invA @ (bdy - Ady @ gamma)
        gammadz = invA @ (bdz - Adz @ gamma)
        fdx, fdy, fdz = _mls_g_3d(gamma, gammadx, gammady, gammadz, B, Bdx, Bdy, Bdz, V)

    if H:
        fdxx, fdyy, fdzz, fdxy, fdxz, fdyz = _mls_H_3d(
            invA,
            bdxx,
            bdyy,
            bdzz,
            bdxy,
            bdxz,
            bdyz,
            Adx,
            Ady,
            Adz,
            Adxx,
            Adyy,
            Adzz,
            Adxy,
            Adxz,
            Adyz,
            gamma,
            gammadx,
            gammady,
            gammadz,
            B,
            Bdx,
            Bdy,
            Bdz,
            Bdxx,
            Bdyy,
            Bdzz,
            Bdxy,
            Bdxz,
            Bdyz,
            V,
        )

    return f, fdx, fdy, fdz, fdxx, fdyy, fdzz, fdxy, fdxz, fdyz


@njit(nogil=True, cache=True)
def _mls_f(
    gamma: ndarray,
    B: ndarray,
    V: ndarray,
) -> ndarray:
    SHP = gamma @ B
    f = SHP.T @ V
    return f


@njit(nogil=True, cache=True)
def _mls_g_1d(
    gamma: ndarray,
    gammadx: ndarray,
    B: ndarray,
    Bdx: ndarray,
    V: ndarray,
) -> ndarray:
    SHPdx = gammadx @ B + gamma @ Bdx
    fdx = SHPdx.T @ V
    return fdx


@njit(nogil=True, cache=True)
def _mls_g_2d(
    gamma: ndarray,
    gammadx: ndarray,
    gammady: ndarray,
    B: ndarray,
    Bdx: ndarray,
    Bdy: ndarray,
    V: ndarray,
) -> Tuple[ndarray, ndarray]:
    SHPdx = gammadx @ B + gamma @ Bdx
    SHPdy = gammady @ B + gamma @ Bdy
    fdx = SHPdx.T @ V
    fdy = SHPdy.T @ V
    return fdx, fdy


@njit(nogil=True, cache=True)
def _mls_g_3d(
    gamma: ndarray,
    gammadx: ndarray,
    gammady: ndarray,
    gammadz: ndarray,
    B: ndarray,
    Bdx: ndarray,
    Bdy: ndarray,
    Bdz: ndarray,
    V: ndarray,
) -> Tuple[ndarray, ndarray, ndarray]:
    SHPdx = gammadx @ B + gamma @ Bdx
    SHPdy = gammady @ B + gamma @ Bdy
    SHPdz = gammadz @ B + gamma @ Bdz
    fdx = SHPdx.T @ V
    fdy = SHPdy.T @ V
    fdz = SHPdz.T @ V
    return fdx, fdy, fdz


@njit(nogil=True, cache=True)
def _mls_H_1d(
    invA: ndarray,
    bdxx: ndarray,
    Adx: ndarray,
    Adxx: ndarray,
    gamma: ndarray,
    gammadx: ndarray,
    B: ndarray,
    Bdx: ndarray,
    Bdxx: ndarray,
    V: ndarray,
) -> ndarray:
    gammadxx = invA @ (bdxx - Adxx @ gamma - 2 * Adx @ gammadx)
    SHPdxx = gammadxx @ B + 2 * gammadx @ Bdx + gamma @ Bdxx
    fdxx = SHPdxx.T @ V
    return fdxx


@njit(nogil=True, cache=True)
def _mls_H_2d(
    invA: ndarray,
    bdxx: ndarray,
    bdyy: ndarray,
    bdxy: ndarray,
    Adx: ndarray,
    Ady: ndarray,
    Adxx: ndarray,
    Adyy: ndarray,
    Adxy: ndarray,
    gamma: ndarray,
    gammadx: ndarray,
    gammady: ndarray,
    B: ndarray,
    Bdx: ndarray,
    Bdy: ndarray,
    Bdxx: ndarray,
    Bdyy: ndarray,
    Bdxy: ndarray,
    V: ndarray,
) -> Tuple[ndarray, ndarray, ndarray]:
    gammadxx = invA @ (bdxx - Adxx @ gamma - 2 * Adx @ gammadx)
    gammadyy = invA @ (bdyy - Adyy @ gamma - 2 * Ady @ gammady)
    gammadxy = invA @ (bdxy - Adxy @ gamma - Adx @ gammady - Ady @ gammadx)
    SHPdxx = gammadxx @ B + 2 * gammadx @ Bdx + gamma @ Bdxx
    SHPdyy = gammadyy @ B + 2 * gammady @ Bdy + gamma @ Bdyy
    SHPdxy = gammadxy @ B + gammadx @ Bdy + gammady @ Bdx + gamma @ Bdxy
    fdxx = SHPdxx.T @ V
    fdyy = SHPdyy.T @ V
    fdxy = SHPdxy.T @ V
    return fdxx, fdyy, fdxy


@njit(nogil=True, cache=True)
def _mls_H_3d(
    invA: ndarray,
    bdxx: ndarray,
    bdyy: ndarray,
    bdzz: ndarray,
    bdxy: ndarray,
    bdxz: ndarray,
    bdyz: ndarray,
    Adx: ndarray,
    Ady: ndarray,
    Adz: ndarray,
    Adxx: ndarray,
    Adyy: ndarray,
    Adzz: ndarray,
    Adxy: ndarray,
    Adxz: ndarray,
    Adyz: ndarray,
    gamma: ndarray,
    gammadx: ndarray,
    gammady: ndarray,
    gammadz: ndarray,
    B: ndarray,
    Bdx: ndarray,
    Bdy: ndarray,
    Bdz: ndarray,
    Bdxx: ndarray,
    Bdyy: ndarray,
    Bdzz: ndarray,
    Bdxy: ndarray,
    Bdxz: ndarray,
    Bdyz: ndarray,
    V: ndarray,
) -> Tuple[ndarray, ndarray, ndarray, ndarray, ndarray, ndarray]:
    gammadxx = invA @ (bdxx - Adxx @ gamma - 2 * Adx @ gammadx)
    gammadyy = invA @ (bdyy - Adyy @ gamma - 2 * Ady @ gammady)
    gammadzz = invA @ (bdzz - Adzz @ gamma - 2 * Adz @ gammadz)
    gammadxy = invA @ (bdxy - Adxy @ gamma - Adx @ gammady - Ady @ gammadx)
    gammadxz = invA @ (bdxz - Adxz @ gamma - Adx @ gammadz - Adz @ gammadx)
    gammadyz = invA @ (bdyz - Adyz @ gamma - Ady @ gammadz - Adz @ gammady)
    SHPdxx = gammadxx @ B + 2 * gammadx @ Bdx + gamma @ Bdxx
    SHPdyy = gammadyy @ B + 2 * gammady @ Bdy + gamma @ Bdyy
    SHPdzz = gammadzz @ B + 2 * gammadz @ Bdz + gamma @ Bdzz
    SHPdxy = gammadxy @ B + gammadx @ Bdy + gammady @ Bdx + gamma @ Bdxy
    SHPdxz = gammadxz @ B + gammadx @ Bdz + gammadz @ Bdx + gamma @ Bdxz
    SHPdyz = gammadyz @ B + gammady @ Bdz + gammadz @ Bdy + gamma @ Bdyz
    fdxx = SHPdxx.T @ V
    fdyy = SHPdyy.T @ V
    fdzz = SHPdzz.T @ V
    fdxy = SHPdxy.T @ V
    fdxz = SHPdxz.T @ V
    fdyz = SHPdyz.T @ V
    return fdxx, fdyy, fdzz, fdxy, fdxz, fdyz



from typing import Union
import numpy as np
import awkward as ak
from numba.core import types as nbtypes, cgutils
from numba.extending import (
    typeof_impl,
    models,
    make_attribute_wrapper,
    register_model,
    box,
    unbox,
    NativeValue,
    overload_method,
)
from scipy.sparse import issparse
from scipy.sparse import csr_matrix as csr_scipy, spmatrix

from .utils import get_shape_sp, _jagged_to_csr_data_, count_cols


__all__ = ["csr_matrix"]


SparseLike = Union[spmatrix, np.ndarray, ak.Array]


class csr_matrix:
    """
    Numba-jittable Python class for a sparse matrices in CSR format.
    The meaning of the input variables is the same as in SciPy, and
    object creation follows the same pattern.

    Parameters
    ----------
    data : SparseLike
        Contains the non-zero values of the matrix, in the order in which
        they would be encountered if we walked along the rows left to
        right and top to bottom. If this is a CSC matrix, the walk
        happens along the columns. From version 0.0.8, `Awkward` arrays
        are also accepted.
        .. versionmodified:: 0.0.8
    indices : numpy.ndarray, Optional
        The indices of the columns (rows) during the walk.
        Default is None.
    indptr : numpy.ndarray, Optional
        Stores row (column) boundaries. Default is None.
    shape : Tuple, Optional
        Default is None.

    Note
    ----
    1) At the moment, this class does not support `NumPy`'s array protocoll.
    If you want this to be the argument to a numpy function, use the
    :func:`to_scipy` method of this class.
    2) The attributed 'data', 'indices', 'indptr' and 'shape' are all
    accessible inside Numba-jitted functions.

    Examples
    --------
    Create from a JaggedArray

    >>> import numpy as np
    >>> from sigmaepsilon.math.linalg.sparse import JaggedArray, csr_matrix
    >>> data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    >>> csr = JaggedArray(data, cuts=[3, 3, 4]).to_csr()
    >>> csr
    3x4 CSR matrix with 10 nonzero values

    You can watch it as a NumPy array

    >>> csr.to_numpy()
    array([[ 1,  2,  3,  0],
           [ 4,  5,  6,  0],
           [ 7,  8,  9, 10]])

    Create from a SciPy sparse matrix

    >>> from scipy.sparse import csr_matrix as csr_scipy
    >>> csr_scipy((3, 4), dtype=np.int8).toarray()
    >>> csr_matrix(scipy_matrix)
    3x4 CSR matrix with 0 nonzero values

    To create the 10 by 10 identity matrix, do this:

    >>> csr_matrix.eye(10)
    10x10 CSR matrix with 10 nonzero values

    You can access rows and row indices of a CSR matrix in Numba
    jitted code, even in 'nopython' mode:

    >>> from numba import jit
    >>> @jit(nopython=True)
    >>> def numba_nopython(csr: csr_matrix, i: int):
    >>>     return csr.row(i), csr.irow(i)
    >>> row = np.array([0, 0, 1, 2, 2, 2])
    >>> col = np.array([0, 2, 2, 0, 1, 2])
    >>> data = np.array([1, 2, 3, 4, 5, 6])
    >>> matrix = csr_scipy((data, (row, col)), shape=(3, 3))
    >>> matrix.toarray()
    array([[1, 0, 2],
           [0, 0, 3],
           [4, 5, 6]], dtype=int32)
    >>> csr = csr_matrix(matrix)
    >>> numba_nopython(csr, 0)
    (array([1., 2.]), array([0, 2]))

    See also
    --------
    :class:`~sigmaepsilon.math.linalg.sparse.jaggedarray.JaggedArray`
    :class:`scipy.sparse.csr_matrix`
    """

    def __init__(
        self,
        data: SparseLike,
        indices: np.ndarray = None,
        indptr: np.ndarray = None,
        shape: tuple = None,
    ):
        if issparse(data):
            data = data.tocsr()
            self.data = data.data.astype(np.float64)
            self.indices = data.indices.astype(np.int32)
            self.indptr = data.indptr.astype(np.int32)
            self.shape = data.shape
        elif isinstance(data, np.ndarray) and indices is None:
            assert (
                len(data.shape) == 2
            ), "If 'data' is a NumPy array, it must be 2 dimensional."
            self.data = data.flatten()
            self.data = self.data.astype(np.float64)
            self.indices = np.tile(np.arange(data.shape[1]), data.shape[0])
            self.indices = self.indices.astype(np.int32)
            self.indptr = np.arange(data.shape[0] + 1) * data.shape[1]
            self.indptr = self.indptr.astype(np.int32)
            self.shape = data.shape
        elif isinstance(data, ak.Array):
            self.data = ak.flatten(data).to_numpy()
            self.data = self.data.astype("float64")
            cc = count_cols(data)
            bi = ak.ArrayBuilder()
            bptr = ak.ArrayBuilder()
            _jagged_to_csr_data_(bi, bptr, cc)
            self.indices = ak.flatten(bi.snapshot()).to_numpy().astype(np.int32)
            self.indptr = ak.flatten(bptr.snapshot()).to_numpy().astype(np.int32)
            self.shape = get_shape_sp(self.indptr)
        else:
            self.data = np.array(data).astype(np.float64)
            self.indices = np.array(indices).astype(np.int32)
            self.indptr = np.array(indptr).astype(np.int32)
            if shape is None:
                shape = get_shape_sp(indptr)
            self.shape = shape

    def to_numpy(self) -> np.ndarray:
        """
        Returns the matrix as a NumPy array.
        .. versionadded:: 0.0.8
        """
        return self.to_scipy().toarray()

    def to_scipy(self) -> csr_scipy:
        """
        Returns data as a `SciPy` object.
        """
        return csr_scipy((self.data, self.indices, self.indptr), shape=self.shape)

    @staticmethod
    def eye(N: int) -> "csr_matrix":
        """
        Returns the NxN identity matrix as a CSR matrix.
        """
        indices = np.arange(N)
        indptr = np.arange(N + 1)
        data = np.ones(N, dtype=float)
        return csr_matrix(data=data, indices=indices, indptr=indptr, shape=(N, N))

    def row(self, i: int = 0) -> np.ndarray:
        """
        Returns the values of the i-th row.

        .. versionmodified:: 0.0.8

        The behavior was changed in version 0.0.8. After that, the
        call only returns the data related to the i-th row. For the
        indices see :func:`irow`.

        .. note::
            This method is available inside Numba-jitted functions,
            even in nopython mode.
        """
        return self.data[self.indptr[i] : self.indptr[i + 1]]

    def irow(self, i: int = 0) -> np.ndarray:
        """
        Returns the colum indices of the values of the i-th row.

        .. versionadded:: 0.0.8

        .. note::
            This method is available inside Numba-jitted functions,
            even in nopython mode.
        """
        return self.indices[self.indptr[i] : self.indptr[i + 1]]

    def __repr__(self):
        N = len(self.data)
        n, m = self.shape
        return f"{n}x{m} CSR matrix of {N} values."


class csr_matrix_nb(nbtypes.Type):  # pragma: no cover
    """Numba type for a sparse matrix."""

    def __init__(self, dtype):
        self.dtype = dtype
        self.data = nbtypes.Array(self.dtype, 1, "C")
        self.indices = nbtypes.Array(nbtypes.int32, 1, "C")
        self.indptr = nbtypes.Array(nbtypes.int32, 1, "C")
        self.shape = nbtypes.UniTuple(nbtypes.int64, 2)
        super(csr_matrix_nb, self).__init__("csr_matrix")


@overload_method(csr_matrix_nb, "row")
def row(csr, i: int):  # pragma: no cover
    if isinstance(csr, csr_matrix_nb):

        def row_impl(csr, i: int):
            return csr.data[csr.indptr[i] : csr.indptr[i + 1]]

        return row_impl


@overload_method(csr_matrix_nb, "irow")
def irow(csr, i: int):  # pragma: no cover
    if isinstance(csr, csr_matrix_nb):

        def irow_impl(csr, i: int):
            return csr.indices[csr.indptr[i] : csr.indptr[i + 1]]

        return irow_impl


@typeof_impl.register(csr_matrix)
def typeof_csr(val, c):  # pragma: no cover
    data = typeof_impl(val.data, c)
    return csr_matrix_nb(data.dtype)


make_attribute_wrapper(csr_matrix_nb, "data", "data")
make_attribute_wrapper(csr_matrix_nb, "indices", "indices")
make_attribute_wrapper(csr_matrix_nb, "indptr", "indptr")
make_attribute_wrapper(csr_matrix_nb, "shape", "shape")


@register_model(csr_matrix_nb)
class csr_model(models.StructModel):  # pragma: no cover
    """Data model for nopython mode."""

    def __init__(self, dmm, fe_type):
        members = [
            ("data", fe_type.data),
            ("indices", fe_type.indices),
            ("indptr", fe_type.indptr),
            ("shape", fe_type.shape),
        ]
        models.StructModel.__init__(self, dmm, fe_type, members)


@unbox(csr_matrix_nb)
def unbox_csr(typ, obj, c):  # pragma: no cover
    """Convert a python object to a numba-native structure."""
    data = c.pyapi.object_getattr_string(obj, "data")
    indices = c.pyapi.object_getattr_string(obj, "indices")
    indptr = c.pyapi.object_getattr_string(obj, "indptr")
    shape = c.pyapi.object_getattr_string(obj, "shape")
    matrix = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    matrix.data = c.unbox(typ.data, data).value
    matrix.indices = c.unbox(typ.indices, indices).value
    matrix.indptr = c.unbox(typ.indptr, indptr).value
    matrix.shape = c.unbox(typ.shape, shape).value
    for att in [data, indices, indptr, shape]:
        c.pyapi.decref(att)
    is_error = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(matrix._getvalue(), is_error=is_error)


@box(csr_matrix_nb)
def box_csr(typ, val, c):  # pragma: no cover
    """Convert a numba-native structure to a python object."""
    matrix = cgutils.create_struct_proxy(typ)(c.context, c.builder, value=val)
    classobj = c.pyapi.unserialize(c.pyapi.serialize_object(csr_matrix))
    data_obj = c.box(typ.data, matrix.data)
    indices_obj = c.box(typ.indices, matrix.indices)
    indptr_obj = c.box(typ.indptr, matrix.indptr)
    shape_obj = c.box(typ.shape, matrix.shape)
    matrix_obj = c.pyapi.call_function_objargs(
        classobj, (data_obj, indices_obj, indptr_obj, shape_obj)
    )
    return matrix_obj



# -*- coding: utf-8 -*-
from .function import Function, FunctionLike
from .relation import Equality, InEquality, Relation
from .variablemanager import VariableManager

__all__ = [
    "Function",
    "FunctionLike",
    "VariableManager",
    "Equality",
    "InEquality",
    "Relation",
]



from typing import Iterable, Union
import numpy as np
from numpy import concatenate as join, ndarray, array_repr
import awkward as ak
from awkward import unflatten, Array as akarray
from numpy.lib.mixins import NDArrayOperatorsMixin
from scipy.sparse import csr_matrix as csr_scipy

from sigmaepsilon.core import Wrapper
from ...arraysetops import unique2d
from .utils import count_cols
from .csr import csr_matrix


__all__ = ["JaggedArray"]


def shape(arr):
    return arr.shape[:2]


def cut(shp):
    return np.full(shp[0], shp[1], dtype=int)


def flatten(arr):
    return arr.flatten()


def get_data(
    data: Iterable, cuts: Iterable = None, fallback: bool = False
) -> Union[ndarray, akarray]:
    if isinstance(data, np.ndarray):
        nD = len(data.shape)
        assert nD <= 2, "Only 1 and 2 dimensional arrays are supported!"
        if nD == 1:
            assert isinstance(cuts, Iterable)
            cuts = np.array(cuts).astype(int)
            data = unflatten(data, cuts)
        return data
    elif isinstance(data, list):
        try:
            if fallback:
                raise NotImplementedError
            # list of lists
            return ak.Array(data)
        except Exception:
            # list of 2d NumPy arrays
            assert all(
                map(lambda arr: len(arr.shape) == 2, data)
            ), "Only 2 dimensional arrays are supported!"
            # NOTE - implementaion 1
            # > Through the back door, but this is probably the cleanest solution of all.
            # > It only requires to create one python list, without further operations on it.
            # NOTE This line is one of the most unreadable things I've ever done.
            data = unflatten(
                join(list(map(flatten, data))), join(list(map(cut, map(shape, data))))
            )
            # NOTE - implementaion 2
            # from operator import add
            # from functools import reduce
            # > This also works, but requires data to jump back and forth just to
            # > have a merged list of lists. It also requires to add nested python lists,
            # > which is probably not the quickest operation in the computational world.
            # data = ak.from_iter(reduce(add, map(lambda arr : ak.Array(arr).to_list(), data)))
            # NOTE - implementaion 3
            # > This also works at creation, but fails later at some operation due to
            # > the specific layout generated by ak.concatenate
            # data = ak.concatenate(list(map(lambda arr : ak.Array(arr), data)))
            return data
    elif isinstance(data, akarray):
        return data
    else:
        if data:
            raise TypeError(f"Invalid type {type(data)}")


HANDLED_FUNCTIONS = {}


class JaggedArray(NDArrayOperatorsMixin, Wrapper):
    """
    A numba-jittable class that handles 2d matrices with a variable
    number of columns per row.

    The class is actually an interface to `awkward.Array`,
    with some additional features, specific to 2d jagged arrays.

    At the moment a JaggedArray can be constructed from
    * a flattened 2d jagged array through unflatteing (we use Awkward here)
    * a list of 1d lists
    * a list of 2d NumPy arrays

    Parameters
    ----------
    data : Iterable
        A 2d numpy array or a list of them.
    cuts : Iterable, Optional
        An iterable that tells how to unflatten an 1d array into a
        2d jagged shape. Only if 'data' is an 1d array. Default is None.
    force_numpy : bool, Optional
        Forces dense inputs to be NumPy arrays in the background.
        Default is True.

    Examples
    --------
    The following defines a dense matrix from a flattened shape:

    >>> import numpy as np
    >>> from sigmaepsilon.math.linalg.sparse import JaggedArray
    >>> data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    >>> JaggedArray(data, cuts=[3, 3, 3])
    array([[1, 2, 3],
           [4, 5, 6],
           [7, 8, 9]])

    Since we use the Awkward library to deflatten arrays, if we provide
    the argument `force_numpy=False`, which is the default behaviour, the data
    is stored as an Awkward array:

    >>> JaggedArray(data, cuts=[3, 3, 3], force_numpy=False)
    JaggedArray([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    The next example defines a jagged array. In this case the data is inevitably
    stored as an Awkward array.

    >>> data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    >>> JaggedArray(data, cuts=[3, 3, 4])
    JaggedArray([[1, 2, 3], [4, 5, 6], [7, 8, 9, 10]])

    >>> JaggedArray([[1, 2], [3, 4, 5]])
    JaggedArray([[1, 2], [3, 4, 5]])

    >>> JaggedArray([np.eye(2), np.eye(3)])
    JaggedArray([[[1, 0], [0, 1]], [[1, 0, 0], [0, 1, 0], [0, 0, 1]]])

    See also
    --------
    :class:`~sigmaepsilon.math.linalg.sparse.csr.csr_matrix`
    :class:`awkward.Array`
    """

    def __init__(
        self,
        data: Iterable = None,
        *,
        cuts: Iterable = None,
        force_numpy: bool = True,
        **kwargs,
    ):
        wrap = kwargs.get("wrap", None)
        if wrap is None:
            assert isinstance(data, Iterable)
            wrap = get_data(data, cuts)
        if isinstance(wrap, akarray) and force_numpy:
            N = len(np.unique(count_cols(wrap)))
            if N == 1:
                wrap = wrap.to_numpy()
        super(JaggedArray, self).__init__(wrap=wrap)

    def __repr__(self):
        if isinstance(self._wrapped, ndarray):
            return array_repr(self._wrapped)
        else:
            return f"{self.__class__.__name__}({self._wrapped})"

    def __array__(self):
        return self._wrapped

    def __getitem__(self, key):
        return self._wrapped.__getitem__(key)

    def __setitem__(self, key, value):
        return self._wrapped.__setitem__(key, value)

    def __len__(self):
        return len(self._wrapped)

    def to_csr(self) -> csr_matrix:
        """
        Returns the topology as a csr_matrix.

        See also
        --------
        :class:`~sigmaepsilon.math.linalg.sparse.csr.csr_matrix`
        """
        return csr_matrix(self._wrapped)

    def to_scipy(self) -> csr_scipy:
        """
        Returns the array as a sparse SciPy CSR matrix.

        See also
        --------
        :class:`scipy.sparse.csr_matrix`
        """
        return self.to_csr().to_scipy()

    def to_ak(self) -> akarray:
        """
        Returns underlying data as an Awkward array.

        See also
        --------
        :class:`awkward.Array`
        """
        if isinstance(self._wrapped, akarray):
            return self._wrapped
        else:
            return akarray(self._wrapped)

    def to_numpy(self) -> ndarray:
        """
        Returns underlying data as a NumPy array. This is only possible
        for regular topologies.
        """
        if isinstance(self._wrapped, ndarray):
            return self.__array__()
        else:
            return self.to_csr().to_numpy()

    def to_array(self) -> Union[akarray, ndarray]:
        """
        Returns the underlying data, which either an Awkward or a NumPy array.
        """
        return self.__array__()

    def to_list(self):
        """
        Returns underlying data as lists.

        .. versionadded:: 0.0.8
        """
        if isinstance(self._wrapped, ndarray):
            return self.__array__().tolist()
        else:
            return self.__array__().to_list()

    def unique(self, *args, **kwargs):
        """
        Returns unique elements, by generalizing the functionality provided
        by :func:`numpy.unique`, see its documentation for the details.
        """
        return np.unique(self, *args, **kwargs)

    def is_jagged(self) -> bool:
        """
        Returns True if the topology is jagged, False otherwise.
        """
        widths = self.widths()
        return not np.all(widths == widths[0])

    def widths(self) -> ndarray:
        """
        Returns the number of columns for each row.
        """
        return count_cols(self._wrapped)

    @property
    def size(self) -> int:
        """
        Number of elements in the array.
        """
        return np.sum(self.widths())

    def flatten(self, return_cuts: bool = False):
        """
        Returns the flattened equivalent of the array.
        """
        if isinstance(self._wrapped, akarray):
            if return_cuts:
                topo = ak.flatten(self._wrapped).to_numpy()
                return self.widths(), topo
            else:
                topo = ak.flatten(self._wrapped).to_numpy()
                return topo
        else:
            if return_cuts:
                return self.widths(), self.to_numpy().flatten()
            else:
                return self.to_numpy().flatten()

    @property
    def shape(self):
        """
        Returns the shape of the data as a tuple. If the topology is
        jagged, the second item is an iterable.

        .. versionadded:: 0.0.8
        """
        if isinstance(self._wrapped, ndarray):
            return self._wrapped.shape
        else:
            return len(self), self.widths()

    def __array_function__(self, func, types, args, kwargs):
        if func not in HANDLED_FUNCTIONS:
            arrs = [arg.to_ak() for arg in args]
            return func(*arrs, **kwargs)
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            if not all(isinstance(x, self.__class__) for x in args[0]):
                raise TypeError("All object must be instances of JaggedArray.")
        else:
            if not all(isinstance(x, self.__class__) for x in args):
                raise TypeError("All object must be instances of JaggedArray.")
        return HANDLED_FUNCTIONS[func](*args, **kwargs)


def implements(numpy_function):
    """
    Register an __array_function__ implementation for JaggedArray
    objects.
    """

    def decorator(func):
        HANDLED_FUNCTIONS[numpy_function] = func
        return func

    return decorator


@implements(np.unique)
def unique(*args, **kwargs):
    return unique2d(args[0]._wrapped, **kwargs)


@implements(np.vstack)
def vstack(*args, **kwargs):
    data = np.concatenate(list(t.flatten() for t in args[0]))
    cuts = np.concatenate(list(t.widths() for t in args[0]))
    return JaggedArray(data, cuts=cuts)



from typing import TypeVar, Callable, Iterable, Union
from collections import OrderedDict

from sympy import Expr, degree, latex
from sympy.core.numbers import One
import numpy as np

from sigmaepsilon.core.kwargtools import getasany

from .metafunction import MetaFunction, substitute


__all__ = ["Function", "FunctionLike"]


FunctionLike = TypeVar("FunctionLike", str, Callable, Expr)


class Function(MetaFunction):
    """
    Base class for all kinds of functions.

    Parameters
    ----------
    f0: Callable
        A callable object that returns function evaluations.
    f1: Callable
        A callable object that returns evaluations of the
        gradient of the function.
    f2: Callable
        A callable object that returns evaluations of the
        Hessian of the function.
    variables: List, Optional
        Symbolic variables. Only required if the function is defined by
        a string or `SymPy` expression.
    value: Callable, Optional
        Same as `f0`.
    gradient: Callable, Optional
        Same as `f1`.
    Hessian: Callable, Optional
        Same as `f2`.
    dimension or dim or d : int, Optional
        The number of dimensions of the domain of the function. Required only when
        going full blind, in most of the cases it can be inferred from other properties.

    Examples
    --------
    >>> from sigmaepsilon.math.function import Function
    >>> import sympy as sy

    Define a symbolic function with positive variables. Note here that if it was not relevant
    from the aspect of the application to indicate that the variables are positive, it
    wouldn't be necessary to explicity specify them using the parameter `variables`.

    >>> variables = ['x1', 'x2', 'x3', 'x4']
    >>> x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    >>> f = Function(3*x1 + 9*x3 + x2 + x4, variables=syms)
    >>> f([0, 6, 0, 4])
    10

    An equivalent definition can be given without using SymPy explicitly:

    >>> f = Function("3*x1 + 9*x3 + x2 + x4", variables=['x1', 'x2', 'x3', 'x4'])
    >>> f([0, 6, 0, 4])
    10

    In both cases, providing the argument 'variables' is optional, but it determines the
    order of the arguments. If the function is defined without the variables being provided,
    it is derived from the input, but the order of the arguments may differ from what you would
    expect.

    >>> f = Function("3*x1 + 9*x3 + x2 + x4")
    >>> f.variables
    (x3, x1, x4, x2)

    Define a numerical function. In this case the dimension of the input must be specified
    explicitly.

    >>> def f0(x, y): return x**2 + y
    >>> def f1(x, y): return np.array([2*x, 1])
    >>> def f2(x, y): return np.array([[0, 0], [0, 0]])
    >>> f = Function(f0, f1, f2, d=2)
    >>> f.linear
    False

    To call the function, call it like you would call the function `f0`:

    >>> f(1, 1)
    2
    >>> f.g(1, 1)
    array([2, 1])

    You can mix different kinds of input signatures:

    >>> def f0(x): return x[0]**2 + x[1]
    >>> def f1(x, y): return np.array([2*x, 1])
    >>> def f2(x, y): return np.array([[0, 0], [0, 0]])
    >>> f = Function(f0, f1, f2, d=2)

    The point is that you always call the resulting `Function` object
    according to your definition. Now your `f0` expects an iterable,
    therefore you can call it like

    >>> f([1, 1])
    2

    but the gradient function expects the same values as two scalars, so
    you call it like this

    >>> f.g(1, 1)
    array([2, 1])

    Explicity defining the variables for a symbolic function is important
    if not all variables appear in the string expression you feed the object with:

    >>> g = Function('3*x + 4*y - 2', variables=['x', 'y', 'z'])
    >>> g.linear
    True

    If you do not specify 'z' as a variable here, the resulting object expects
    two values

    >>> h = Function('3*x + 4*y - 2')
    >>> h([1, 1])
    5

    The variables can be `SymPy` variables as well:

    >>> m = Function('3*x + 4*y - 2', variables=sy.symbols('x y z'))
    >>> m.linear
    True
    >>> m([1, 2, -30])
    9
    """

    # NOTE domain is missing from the possible parameters
    # NOTE investigate if dimensions could be derived

    def __init__(
        self,
        f0: FunctionLike | None = None,
        f1: Callable | None = None,
        f2: Callable | None = None,
        *args,
        variables: Iterable | None = None,
        **kwargs
    ):
        super().__init__()
        self.update(f0, f1, f2, *args, variables=variables, **kwargs)

    def update(
        self,
        f0: FunctionLike | None = None,
        f1: Callable | None = None,
        f2: Callable | None = None,
        *_,
        variables: Iterable | None = None,
        **kwargs
    ):
        self.from_str = None
        if f0 is not None:
            if isinstance(f0, str):
                kwargs.update(self._str_to_func(f0, variables=variables, **kwargs))
                self.from_str = True
            elif isinstance(f0, Expr):
                kwargs.update(self._sympy_to_func(f0, variables=variables, **kwargs))
        self.expr = kwargs.get("expr", None)
        self.variables = kwargs.get("variables", variables)
        self.f0 = kwargs.get("value", f0)
        self.f1 = kwargs.get("gradient", f1)
        self.f2 = kwargs.get("Hessian", f2)
        self.dimension = getasany(["d", "dimension", "dim"], None, **kwargs)
        self.domain = kwargs.get("domain", None)
        self.vmap = kwargs.get("vmap", None)

    @property
    def symbolic(self) -> bool:
        """
        Returns True if the function is a fit subject of symbolic manipulation.
        This is probably only true if the object was created from a string or
        `sympy` expression.
        """
        return self.expr is not None

    @property
    def linear(self) -> bool:
        """
        Returns True if the function is at most linear in all of its variables.
        """
        if self.symbolic:
            return all(
                np.array([degree(self.expr, v) for v in self.variables], dtype=int) <= 1
            )
        else:
            return self.f2 is None

    def linear_coefficients(self, normalize: bool = False) -> Union[Iterable, None]:
        """
        Returns the linear coeffiecients, if the function is symbolic.
        """
        d = self.coefficients(normalize)
        if d:
            return {
                key: value for key, value in d.items() if len(key.free_symbols) <= 1
            }
        return None

    def coefficients(self, normalize: bool = False):
        """
        Returns the coefficients if the function is symbolic.
        """
        try:
            d = OrderedDict({x: 0 for x in self.variables})
            d.update(self.expr.as_coefficients_dict())
            if not normalize:
                return d
            else:
                res = OrderedDict()
                for key, value in d.items():
                    if len(key.free_symbols) == 0:
                        res[One()] = value * key
                    else:
                        res[key] = value
                return res
        except Exception:
            return None

    def to_latex(self) -> str:
        """
        Returns the LaTeX code of the symbolic expression of the object.
        Only for simbolic functions.
        """
        if self.symbolic:
            return latex(self.expr)
        else:
            raise TypeError("This is exclusive to symbolic functions.")

    def subs(
        self, values: Iterable, variables: Iterable | None = None, inplace: bool = False
    ) -> "Function":
        """
        Substitites values for variables.
        """
        if not self.symbolic:
            raise TypeError("This is exclusive to symbolic functions.")

        expr = substitute(self.expr, values, variables, as_string=self.from_str)
        kwargs = self._sympy_to_func(expr=expr, variables=variables)

        if not inplace:
            return Function(None, None, None, **kwargs)
        else:
            self.update(None, None, None, **kwargs)
            return self



from abc import abstractmethod
import weakref
from typing import Tuple, Union
import numbers
from functools import partial
from copy import copy, deepcopy

import numpy as np
from numpy import array_repr, array_str, ndarray
from numpy.lib.mixins import NDArrayOperatorsMixin

from sigmaepsilon.core import Wrapper
from sigmaepsilon.core.abstract import ABC_Safe

from ..utils import ascont, minmax
from .exceptions import LinalgMissingInputError


__all__ = ["ArrayWrapper", "ArrayLike", "TensorLike", "FrameLike"]


class Array(ABC_Safe, ndarray):
    """
    Base backend class for array-like classes. Although you don't really need
    to directly create instances of this class, you can use it like if it was
    a ``numpy.ndarray`` instance.

    The class has a strong metaclass, which means that there is a safety mechanism
    that prevents you from unintentionally crashing the internal behaviour of the
    class upon subclassing. This practically means, that you will see an error if
    you try to shadow a definition in any of the base classes of the class. For this
    reason it is safer to subclass this class rather than to directly subclass
    NumPy's ndarray class.

    See also
    --------
    :class:`~numpy.ndarray`
    :class:`~sigmaepsilon.core.abstract.ABC_Safe`
    """

    def __new__(
        subtype,
        shape=None,
        dtype=float,
        buffer=None,
        offset=0,
        strides=None,
        order=None,
        frame=None,
    ):
        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments. This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to InfoArray.__array_finalize__
        obj = super().__new__(subtype, shape, dtype, buffer, offset, strides, order)
        obj._frame = frame
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(InfoArray, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. InfoArray():
        #    obj is None
        #    (we're in the middle of the InfoArray.__new__
        #    constructor, and self.info will be set when we return 0to
        #    InfoArray.__new__)
        if obj is None:
            return
        # From view casting - e.g arr.view(InfoArray):
        #    obj is arr
        #    (type(obj) can be InfoArray)
        # From new-from-template - e.g infoarr[:3]
        #    type(obj) is InfoArray
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'info', because this
        # method sees all creation of default objects - with the
        # InfoArray.__new__ constructor, but also with
        # arr.view(InfoArray).
        self._frame = getattr(obj, "_frame", None)

    @property
    def frame(self) -> "FrameLike":
        """
        Returns the frame of the vector.
        """
        return self._frame

    @frame.setter
    def frame(self, value: "FrameLike"):
        """
        Sets the frame.
        """
        if isinstance(value, FrameLike):
            self._frame = value
        else:
            raise TypeError("Value must be a {} instance".format(FrameLike))

    def __repr__(self):
        return array_repr(self)

    def __str__(self):
        return array_str(self)


class ArrayWrapper(NDArrayOperatorsMixin, Wrapper):
    """
    Base frontend class for array-like classes. Use it like if it
    was a ``numpy.ndarray`` instance.
    """

    _array_cls_ = Array

    # One might also consider adding the built-in list type to this
    # list, to support operations like np.add(array_like, list)
    _HANDLED_TYPES_ = (np.ndarray, numbers.Number, list)

    def __init__(self, *args, cls_params=None, contiguous: bool = True, **kwargs):
        if len(args) > 0 and isinstance(args[0], np.ndarray):
            buf = ascont(args[0]) if contiguous else args[0]
        else:
            buf = np.array(*args, **kwargs)
        cls_params = dict() if cls_params is None else cls_params
        buf = np.array(buf, dtype=float)
        self._array = self._array_cls_(
            shape=buf.shape, buffer=buf, dtype=buf.dtype, **cls_params
        )
        super(ArrayWrapper, self).__init__(wrap=self._array)

    @property
    def dim(self) -> int:
        """
        Returns the dimension of the array.
        """
        return len(self._array.shape)

    @property
    def minmax(self) -> Tuple[float]:
        """
        Returns the minimum and maximum values of the array.
        """
        return minmax(self._array)

    def chop(self, tol: float = 1e-12) -> "ArrayWrapper":
        """
        Sets very small values (in an absolute sense) to zero.

        .. versionadded:: 1.0.5

        Parameters
        ----------
        tol: float, Optional
            The values whose absolute value is less than this limit are
            set to zero. Default is 1e-12.

        Returns
        -------
        ~`sigmaepsilon.math.linalg.meta.ArrayWrapper`
            The object the call was made upon.
        """
        self._array[np.where(np.abs(self._array) < tol)] = 0.0
        return self

    def __array__(self, dtype=None):
        if dtype is not None:
            return self._array.astype(dtype)
        return self._array

    def __getitem__(self, key):
        return self._array.__getitem__(key)

    def __setitem__(self, key, value):
        return self._array.__setitem__(key, value)

    def __len__(self):
        return self._array.shape[0]

    def to_numpy(self) -> np.ndarray:
        """
        Returns the data as a pure NumPy array.
        """
        return self.__array__()

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = kwargs.get("out", ())
        for x in inputs + out:
            # Only support operations with instances of _HANDLED_TYPES.
            # Use ArrayLike instead of type(self) for isinstance to
            # allow subclasses that don't override __array_ufunc__ to
            # handle ArrayLike objects.
            if not isinstance(x, self._HANDLED_TYPES_ + (Array, ArrayWrapper)):
                raise TypeError(f"Invalid type encountered at {ufunc}")

        # Defer to the implementation of the ufunc on unwrapped values.
        inputs = tuple(x._array if isinstance(x, ArrayWrapper) else x for x in inputs)
        if out:
            kwargs["out"] = tuple(
                x._array if isinstance(x, ArrayWrapper) else x for x in out
            )
        result = getattr(ufunc, method)(*inputs, **kwargs)

        if type(result) is tuple:
            # multiple return values
            return tuple(type(self)(x) for x in result)
        elif method == "at":
            # no return value
            return None
        else:
            # one return value
            return type(self)(result)

    def __repr__(self):
        return array_repr(self)

    def __str__(self):
        return array_str(self)


class FrameLike(ArrayWrapper):
    """
    Base class for reference frames.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._weakrefs = weakref.WeakValueDictionary()

    @abstractmethod
    def dcm(self) -> ndarray: ...

    @abstractmethod
    def show(self) -> ndarray: ...

    @abstractmethod
    def orient(self) -> "FrameLike": ...

    @abstractmethod
    def orient_new(self) -> "FrameLike": ...

    @abstractmethod
    def Gram(self) -> ndarray: ...

    @abstractmethod
    def dual(self) -> "FrameLike": ...

    @abstractmethod
    def transpose(self, inplace: bool = False) -> "FrameLike": ...

    @property
    def T(self) -> "FrameLike":
        """
        Returns the transpose.
        """
        return self.transpose(inplace=False)

    def _register_tensorial_(self, v: "TensorLike"):
        """
        Registers tensorial objects by appending a weak reference to the set
        of weak references. Registered objects change their components upon
        changes of their supporting frame.
        """
        self._weakrefs[id(v)] = v

    def _unregister_tensorial_(self, v: "TensorLike") -> bool:
        """
        Unregisters previously registered tensorial objects. Returns True if the
        object was found in the registry, False if it was not.
        """
        k = id(v)
        if k in self._weakrefs:
            del self._weakrefs[k]
            return True
        return False


class TensorLike(ArrayWrapper):
    """
    Abstract base class for numerical data classes that walk and talk like
    a tensor does.
    """

    _frame_cls_: FrameLike = None
    _rank_: int = None

    def __init__(
        self,
        *args,
        frame: FrameLike = None,
        bulk: bool = None,
        rank: int = None,
        **kwargs,
    ):
        if len(args) > 0 and isinstance(args[0], np.ndarray):
            if not self._verify_input(args[0], bulk=bulk, rank=rank):
                raise ValueError("Invalid input to Tensor class.")

        cls_params = kwargs.get("cls_params", dict())
        if frame is not None:
            if not isinstance(frame, FrameLike):
                raise TypeError(f"The frame must be of type {FrameLike}.")

            cls_params["frame"] = frame
        else:
            if not (len(args) > 0 and isinstance(args[0], np.ndarray)):
                raise LinalgMissingInputError(
                    "A frame or an array of components is required."
                )

            arr = args[0]
            if bulk:
                frame = self.__class__._frame_cls_(dim=arr.shape[1])
            else:
                frame = self.__class__._frame_cls_(dim=arr.shape[0])
            cls_params["frame"] = frame

        kwargs["cls_params"] = cls_params
        super().__init__(*args, **kwargs)

        if self._array._frame is None:
            self._array._frame = self.__class__._frame_cls_(dim=self._array.shape)

        self.frame._register_tensorial_(self)

        self._bulk = bulk

        if rank is not None:
            if self.__class__._rank_ is not None:
                raise ValueError("Rank is already defined on the class level.")
            else:
                self._rank = rank
        else:
            self._rank = None

    def __deepcopy__(self, memo):
        return self.__copy__(memo)

    def __copy__(self, memo=None):
        cls = type(self)
        copy_function = copy if (memo is None) else partial(deepcopy, memo=memo)
        is_deep = memo is not None
        frame_cls = cls._frame_cls_

        f = self.frame
        if is_deep:
            ax = deepcopy(f.axes)
            memo[id(f.axes)] = ax
            frame = frame_cls(ax)
        else:
            frame = f

        arr = copy_function(self.array)
        if is_deep:
            memo[id(self.array)] = arr

        return cls(arr, frame=frame)

    @classmethod
    def _from_any_input(cls, *args, **kwargs) -> "TensorLike":
        raise NotImplementedError

    @classmethod
    def _verify_input(cls, arr: ndarray, *_, **kwargs) -> bool:
        raise NotImplementedError

    @property
    def rank(self) -> int:
        """
        Returns the tensor rank (or order).
        """
        if self._rank:
            return self._rank
        else:
            if self.__class__._rank_:
                return self.__class__._rank_
            else:
                if self._bulk:
                    return len(self.array.shape) - 1
                else:
                    return len(self.array.shape)

    @property
    def array(self) -> Array:
        """
        Returns the coordinates of the vector.
        """
        return self._array

    @array.setter
    def array(self, value: np.ndarray):
        """
        Sets the coordinates of the vector.
        """
        array = np.array(value)
        assert array.shape == self._array.shape
        self._array[...] = array

    @property
    def frame(self) -> FrameLike:
        """
        Returns the frame of the vector.
        """
        return self.array.frame

    @frame.setter
    def frame(self, value: FrameLike):
        """
        Sets the frame of the vector.

        Note
        ----
        Setting a new frame may change the compoenents of the instance.
        """
        if isinstance(value, FrameLike):
            f = value
        elif isinstance(value, ndarray):
            f = self._frame_cls_(value)
        else:
            raise TypeError(f"Value must be a {ndarray} or a {FrameLike} instance")
        array = self.show(f)
        self.array.frame._unregister_tensorial_(self)
        self.array = array
        self.array.frame = f
        f._register_tensorial_(self)

    @property
    def T(self) -> "TensorLike":
        """
        Returns the transpose.
        """
        return self.transpose(inplace=False)

    def transpose(self, inplace: bool = False) -> "TensorLike":
        """
        Either transposes the array of the tensor, or returns a copy
        of it with the components transposed.

        Parameters
        ----------
        inplace: bool, Optional
            If ``True``, the operation is performed on the instance the call
            is made upon. Default is False.

        Note
        ----
        The rule of transposition differs from the one implemented in NumPy, as
        only tensorial axes are being transposed.
        """
        r = self.rank
        shape = self.array.shape
        indices = tuple(range(len(shape)))
        data_indices = indices[:-r]
        tensor_indices = indices[len(shape) - r :]
        indices = data_indices + tensor_indices[::-1]
        if inplace:
            self._array = np.transpose(self.array, indices)
            return self
        else:
            f = self.frame
            frame = f.__class__(np.copy(f.axes))
            return self.__class__(np.transpose(self.array, indices), frame=frame)

    @abstractmethod
    def show(self) -> Array: ...

    @abstractmethod
    def orient(self) -> "TensorLike": ...

    @abstractmethod
    def orient_new(self) -> "TensorLike": ...

    def is_bulk(self):
        """
        Returns True if the object represents a collection of tensors, False
        otherwise.
        """
        if self._bulk:
            return True
        else:
            not self.rank == len(self.array.shape)


ArrayLike = Union[ArrayWrapper, ndarray, Array]



from typing import Union, Iterable
import numbers
import itertools

import numpy as np
from numpy import ndarray
from numba import njit, prange, guvectorize as guv
import sympy as sy
from sympy import symbols, Matrix
from sympy.physics.vector import ReferenceFrame as SymPyFrame

from sigmaepsilon.core.alphabet import latinrange

from .meta import TensorLike, ArrayWrapper, FrameLike
from .exceptions import LinalgOperationInputError, LinalgMissingInputError, LinalgError

__cache = True


__all__ = [
    "permutation_tensor",
    "dot",
    "cross",
    "normalize_frame",
    "Gram",
    "dual_frame",
    "random_pos_semidef_matrix",
    "random_posdef_matrix",
    "inv_sym_3x3",
    "vpath",
    "det3x3",
    "det2x2",
    "inv2x2",
    "inv2x2u",
    "adj3x3",
    "inv3x3u",
    "inv3x3",
    "inv3x3_bulk",
    "inv3x3_bulk2",
    "normalize",
    "normalize2d",
    "norm",
    "norm2d",
    "linspace",
    "linspace1d",
    "inv",
    "show_vector",
    "show_frame",
    "rotation_matrix",
    "generalized_left_inverse",
    "generalized_right_inverse",
    "generalized_inverse",
]


def rotation_matrix(
    rot_type: str, amounts: Iterable, rot_order: Union[str, int] = ""
) -> ndarray:
    """
    Returns a rotation matrix using the mechanism provided by
    `sympy.physics.vector.ReferenceFrame.orientnew`.

    Parameters
    ----------
    rot_type: str
        The method used to generate the direction cosine matrix. Supported
        methods are:

        - ``'Axis'``: simple rotations about a single common axis
        - ``'DCM'``: for setting the direction cosine matrix directly
        - ``'Body'``: three successive rotations about new intermediate
            axes, also called "Euler and Tait-Bryan angles"
        - ``'Space'``: three successive rotations about the parent
            frames' unit vectors
        - ``'Quaternion'``: rotations defined by four parameters which
            result in a singularity free direction cosine matrix

    amounts: Iterable
        Expressions defining the rotation angles or direction cosine
        matrix. These must match the ``rot_type``. See examples below for
        details. The input types are:

        - ``'Axis'``: 2-tuple (expr/sym/func, Vector)
        - ``'DCM'``: Matrix, shape(3, 3)
        - ``'Body'``: 3-tuple of expressions, symbols, or functions
        - ``'Space'``: 3-tuple of expressions, symbols, or functions
        - ``'Quaternion'``: 4-tuple of expressions, symbols, or
            functions

    rot_order: str or int, Optional
        If applicable, the order of the successive of rotations. The string
        ``'123'`` and integer ``123`` are equivalent, for example. Required
        for ``'Body'`` and ``'Space'``.

    Returns
    -------
    ReferenceFrame
        A new ReferenceFrame object.

    See Also
    --------
    :func:`sympy.physics.vector.ReferenceFrame.orientnew`

    Example
    -------
    Define a standard Cartesian frame and rotate it around axis 'Z'
    with 180 degrees:

    >>> from sigmaepsilon.math.linalg.utils import rotation_matrix
    >>> R = rotation_matrix('Space', [0, 0, np.pi], 'XYZ')
    """
    source = SymPyFrame("S")
    target = source.orientnew("T", rot_type, amounts, rot_order)
    dcm = np.array(target.dcm(source).evalf()).astype(float)
    return dcm


def permutation_tensor(dim: int = 3) -> ndarray:
    """
    Returns the Levi-Civita pseudotensor for N dimensions.

    Parameters
    ----------
    N : int, Optional
        The number of dimensions. Default is 3.
    """
    arr = np.zeros(tuple([dim for _ in range(dim)]))
    mat = np.zeros((dim, dim), dtype=np.int32)
    for x in itertools.permutations(tuple(range(dim))):
        mat[:, :] = 0
        for i, j in zip(range(dim), x):
            mat[i, j] = 1
        arr[x] = int(np.linalg.det(mat))
    return arr


def dot(
    a: Union[TensorLike, ArrayWrapper],
    b: Union[TensorLike, ArrayWrapper],
    out: Union[TensorLike, ArrayWrapper] = None,
    frame: FrameLike = None,
    axes: Union[list, tuple] = None,
) -> Union[TensorLike, ndarray, numbers.Number]:
    """
    Returns the dot product (without complex conjugation) of two quantities. The behaviour
    coincides with NumPy when all inputs are arrays and generalizes when they are not,
    but all inputs must be either all arrays or all tensors of some kind. The operation for
    tensors of order 1 and 2 have dedicated implementations, for higher order tensors
    it generalizes to tensor contraction along specified axes.

    Parameters
    ----------
    a: :class:`~sigmaepsilon.math.linalg.meta.TensorLike` or ArrayLike
       A tensor or an array.
    b: :class:`~sigmaepsilon.math.linalg.meta.TensorLike` or ArrayLike
       A tensor or an array.
    out: ArrayLike, Optional
        Output argument. This must have the exact kind that would be returned if it was
        not used. See `numpy.dot` for the details. Only if all inputs are ArrayLike.
        Default is None.
    frame: FrameLike, Optional
        The target frame of the output. Only if all inputs are TensorLike. If not specified,
        the returned tensor migh be returned in an arbitrary frame, depending on the inputs.
        Default is None.
    axes: tuple or list, Optional
        The indices along which contraction happens if any of the input tensors have a rank
        higher than 2. Default is None.

    Returns
    -------
    :class:`~sigmaepsilon.math.linalg.meta.TensorLike` or numpy.ndarray or scalar
        An array or a tensor, depending on the inputs.

    Notes
    -----
    For general tensors, the current implementation has an upper limit considering the rank
    of the input tensors. The sum of the ranks of the input tensors plus the sum of contraction
    indices must be at most 26.

    References
    ----------
    https://mathworld.wolfram.com/DotProduct.html

    Examples
    --------
    When working with NumPy arrays, the behaviour coincides with `numpy.dot`. To take the dot
    product of a 2nd order tensor and a vector, use it like this:

    >>> from sigmaepsilon.math.linalg import ReferenceFrame, Vector, Tensor2
    >>> from sigmaepsilon.math.linalg import dot
    >>> frame = ReferenceFrame(np.eye(3))
    >>> A = Tensor2(np.eye(3), frame=frame)
    >>> v = Vector(np.array([1., 0, 0]), frame=frame)
    >>> dot(A, v)
    Array([1., 0., 0.])

    For general tensors, you have to specify the axes along which contraction happens:

    >>> from sigmaepsilon.math.linalg import Tensor
    >>> A = Tensor(np.ones((3, 3, 3, 3)), frame=frame)  # a tensor of order 4
    >>> B = Tensor(np.ones((3, 3, 3)), frame=frame)  # a tensor of order 3
    >>> dot(A, B, axes=(0, 0)).rank
    5
    """
    if isinstance(a, TensorLike) and isinstance(b, TensorLike):
        ra, rb = a.rank, b.rank
        result = None
        if ra == 1 and rb == 1:
            if out is not None:
                raise LinalgOperationInputError(
                    "Parameter 'out' is not allowed with tensors."
                )
            return np.dot(a.show(), b.show())
        elif ra == 2 and rb == 1:
            arr = (a.array @ b.show(a.frame.dual()).T).T
            result = b.__class__(arr, frame=a.frame)
        elif ra == 1 and rb == 2:
            arr = (a.array.T @ b.show(a.frame.dual()).T).T
            result = a.__class__(arr, frame=a.frame)
        elif ra == 2 and rb == 2:
            g = a.frame.Gram()
            result = a.__class__(a.array @ g @ b.show(a.frame), frame=a.frame)
        else:
            if not axes:
                msg = "The parameter 'axes' is required for tensor contraction of general tensors."
                raise LinalgMissingInputError(msg)
            ia = latinrange(ra, start=ord("a"))
            ib = latinrange(rb, start=ord("a") + ra)
            ax_a, ax_b = axes
            ic = latinrange(1, start=ord("a") + ra + rb)[0]
            ia[ax_a] = ic
            ib[ax_b] = ic
            command = "..." + "".join(ia) + "," + "..." + "".join(ib)
            arr = np.einsum(command, a.show(), b.show(), optimize="greedy")
            result = a.__class__._from_any_input(arr)
        if frame:
            result.frame = frame
        return result
    if frame:
        raise LinalgOperationInputError(
            "Parameter 'frame' is exclusive for tensorial inputs."
        )
    if not all([isinstance(x, (ndarray, ArrayWrapper, list)) for x in [a, b]]):
        raise TypeError("Invalid types encountered for dot product.")
    inputs = [x._array if isinstance(x, ArrayWrapper) else x for x in [a, b]]
    return np.dot(*inputs, out=out)


def cross(
    a: Union[TensorLike, ArrayWrapper],
    b: Union[TensorLike, ArrayWrapper],
    *args,
    frame: FrameLike = None,
    **kwargs,
) -> Union[TensorLike, ndarray]:
    """
    Calculates the cross product of two vectors or one vector and a second order
    tensor. The behaviour coincides with NumPy when all inputs are arrays and generalizes
    when they are not, but all inputs must be either all arrays or all tensors of some kind.

    Parameters
    ----------
    *args : Tuple, Optional
        Positional arguments forwarded to NumPy, if all input objects are arrays.
    a: :class:`~sigmaepsilon.math.linalg.meta.TensorLike` or ArrayLike
        A tensor or an array.
    b: :class:`~sigmaepsilon.math.linalg.meta.TensorLike` or ArrayLike
        A tensor or an array.
    frame: FrameLike, Optional
        The target frame of the output. Only if all inputs are TensorLike. If not specified,
        the returned tensor migh be returned in an arbitrary frame, depending on the inputs.
        Default is None.
    **kwargs: dict, Optional
        Keyword arguments forwarded to `numpy.cross`. As of NumPy version '1.22.4', there
        are no keyword arguments for `numpy.cross`, this is to assure compliance with
        all future versions of numpy.

    Returns
    -------
    numpy.ndarray or :class:`~sigmaepsilon.math.linalg.meta.TensorLike`
        An 1d or 2d array, or an 1d or 2d tensor, depending on the inputs.

    References
    ----------
    https://mathworld.wolfram.com/CrossProduct.html

    Examples
    --------
    The cross product of two vectors results in a vector:

    >>> from sigmaepsilon.math.linalg import ReferenceFrame, Vector, Tensor2
    >>> from sigmaepsilon.math.linalg import cross
    >>> frame = ReferenceFrame(np.eye(3))
    >>> a = Vector(np.array([1., 0, 0]), frame=frame)
    >>> b = Vector(np.array([0, 1., 0]), frame=frame)
    >>> cross(a, b)
    Array([0., 0., 1.])

    The cross product of a second order tensor and a vector result a second order tensor:

    >>> A = Tensor2(np.eye(3), frame=frame)
    >>> cross(A, b)
    Array([[ 0.,  0., -1.],
           [ 0.,  0.,  0.],
           [ 1.,  0.,  0.]])
    """
    if isinstance(a, TensorLike) and isinstance(b, TensorLike):
        ra, rb = a.rank, b.rank
        result = None
        if ra == 1 and rb == 1:
            arr = np.cross(a.array, b.show(a.frame), axis=0)
            result = a.__class__(arr, frame=a.frame)
        elif ra == 2 and rb == 1:
            arr = np.cross(a.show(), b.show(), axis=0)
            result = a.__class__(arr)
        elif ra == 1 and rb == 2:
            arr = np.cross(a.show(), b.show(), axis=0)
            result = b.__class__(arr)
        else:
            msg = (
                "The cross product is not implemented",
                f"for tensors of rank {ra} and {rb}",
            )
            raise NotImplementedError(msg)
        if frame:
            result.frame = frame
        return result
    if frame:
        raise LinalgOperationInputError(
            "Parameter 'frame' is exclusive for tensorial inputs."
        )
    if any([isinstance(x, TensorLike) for x in [a, b]]):
        raise TypeError("Invalid types encountered for dot product.")
    if not all([isinstance(x, (ndarray, ArrayWrapper, list)) for x in [a, b]]):
        raise TypeError("Invalid types encountered for dot product.")
    inputs = [x._array if isinstance(x, ArrayWrapper) else x for x in [a, b]]
    return np.cross(*inputs, *args, **kwargs)


def show_vector(dcm: ndarray, arr: ndarray) -> ndarray:
    """
    Returns the coordinates of a single or multiple vectors in a frame specified
    by one or several DCM matrices. The function can handle the following scenarios:

        - a single (1d) vector and a single (2d) dcm matrix (trivial case)
        - a stack of vectors (2d) and a single (2d) dcm matrix
        - a stack of fectors (2d) and dcm matrices for each vector in the stack (3d)

    .. versionadded:: 1.0.5

    Parameters
    ----------
    dcm: numpy.ndarray
        The dcm matrix of the transformation as a 2d or 3d float array.
    arr: numpy.ndarray
        1d or 2d float array of coordinates of a single vector. If it is 2d, then
        it is assumed that the coordinates of the i-th vector are accessible as
        `arr[i]`.

    Returns
    -------
    numpy.ndarray
        The new coordinates with the same shape as `arr`.
    """
    if len(arr.shape) == 1 and len(dcm.shape) == 2:
        return _show_vector(dcm, arr)  # dcm @ arr
    elif len(arr.shape) == 2 and len(dcm.shape) == 2:
        return _show_vectors(dcm, arr)  # dcm @ arr
    elif len(arr.shape) == 2 and len(dcm.shape) == 3:
        return _show_vectors_multi(dcm, arr)  # dcm @ arr
    else:
        msg = (
            "Mismatch in shapes!"
            f"Input one has shape {dcm.shape} and input two has shape {arr.shape}."
            "See the docs for the correct input shapes."
        )
        raise LinalgOperationInputError(msg)


def show_frame(dcm: ndarray, arr: ndarray) -> ndarray:
    if len(arr.shape) == 2 and len(dcm.shape) == 2:
        return _show_frame(dcm, arr)  # dcm @ arr
    elif len(arr.shape) == 3 and len(dcm.shape) == 2:
        return _show_frames(dcm, arr)  # dcm @ arr
    elif len(arr.shape) == 3 and len(dcm.shape) == 3:
        return _show_frames_multi(dcm, arr)  # dcm @ arr
    else:
        msg = (
            "Mismatch in shapes!"
            f"Input one has shape {dcm.shape} and input two has shape {arr.shape}."
            "See the docs for the correct input shapes."
        )
        raise LinalgOperationInputError(msg)


@njit(nogil=True, cache=__cache)
def _show_vector(dcm: ndarray, arr: ndarray) -> ndarray:
    """
    Returns the coordinates of a single vector in a frame specified
    by a DCM matrix.

    Parameters
    ----------
    dcm: numpy.ndarray
        The dcm matrix of the transformation as a 2d float array.
    arr: numpy.ndarray
        1d float array of coordinates of a single vector.

    Returns
    -------
    numpy.ndarray
        The new coordinates of the vector with the same shape as `arr`.
    """
    return dcm @ arr


@njit(nogil=True, parallel=True, cache=__cache)
def _show_vectors(dcm: ndarray, arr: ndarray) -> ndarray:
    """
    Returns the coordinates of multiple vectors in a frame specified
    by a DCM matrix.

    Parameters
    ----------
    dcm: numpy.ndarray
        The dcm matrix of the transformation as a 2d float array.
    arr: numpy.ndarray
        2d float array of coordinates of multiple vectors.

    Returns
    -------
    numpy.ndarray
        The new coordinates of the vectors with the same shape as `arr`.
    """
    res = np.zeros_like(arr)
    for i in prange(arr.shape[0]):
        res[i] = dcm @ arr[i, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _show_vectors_multi(dcm: ndarray, arr: ndarray) -> ndarray:
    """
    Returns the coordinates of multiple vectors and multiple DCM matrices.

    Parameters
    ----------
    dcm: numpy.ndarray
        The dcm matrix of the transformation as a 3d float array.
    arr: numpy.ndarray
        2d float array of coordinates of multiple vectors.

    Returns
    -------
    numpy.ndarray
        The new coordinates of the vectors with the same shape as `arr`.
    """
    res = np.zeros_like(arr)
    for i in prange(arr.shape[0]):
        res[i] = dcm[i] @ arr[i, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _show_frame(dcm: ndarray, arr: ndarray) -> ndarray:
    """
    Returns the coordinates of a single frame in a target frame specified
    by a DCM matrix.

    Parameters
    ----------
    dcm: numpy.ndarray
        The dcm matrix of the transformation as a 2d float array.
    arr: numpy.ndarray
        2d float array of coordinates of a single frame.

    Returns
    -------
    numpy.ndarray
        The new coordinates of the frame with the same shape as `arr`.
    """
    res = np.zeros_like(arr)
    for i in prange(arr.shape[-1]):
        res[i, :] = dcm @ arr[i, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _show_frames(dcm: ndarray, arr: ndarray) -> ndarray:
    """
    Returns the coordinates of multiple frames in a target frame specified
    by a DCM matrix.

    Parameters
    ----------
    dcm: numpy.ndarray
        The dcm matrix of the transformation as a 2d float array.
    arr: numpy.ndarray
        3d float array of coordinates of multiple frames.

    Returns
    -------
    numpy.ndarray
        The new coordinates of the frames with the same shape as `arr`.
    """
    res = np.zeros_like(arr)
    for i in prange(arr.shape[0]):
        for j in prange(arr.shape[-1]):
            res[i, j, :] = dcm @ arr[i, j, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _show_frames_multi(dcm: ndarray, arr: ndarray) -> ndarray:
    """
    Returns the coordinates of multiple frames and multiple DCM matrices.

    Parameters
    ----------
    dcm: numpy.ndarray
        The dcm matrix of the transformation as a 3d float array.
    arr: numpy.ndarray
        3d float array of coordinates of multiple frames.

    Returns
    -------
    numpy.ndarray
        The new coordinates of the frames with the same shape as `arr`.
    """
    res = np.zeros_like(arr)
    for i in prange(arr.shape[0]):
        for j in prange(arr.shape[-1]):
            res[i, j, :] = dcm[i] @ arr[i, j, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _transpose_multi(dcm: ndarray) -> ndarray:
    N = dcm.shape[0]
    res = np.zeros_like(dcm)
    for i in prange(N):
        res[i, :, :] = dcm[i].T
    return res


def transpose_axes(arr: ndarray) -> ndarray:
    if len(arr.shape) == 2:
        return arr.T
    elif len(arr.shape) == 3:
        # FIXME this might be unnecessary
        return _transpose_multi(arr)
    else:
        shape = arr.shape
        indices = tuple(range(len(shape)))
        data_indices = indices[:-2]
        tensor_indices = indices[len(shape) - 2 :]
        indices = data_indices + tensor_indices[::-1]
        return np.transpose(arr, indices)


def normalize_frame(axes: ndarray) -> ndarray:
    """
    Returns the frame with normalized base vectors.

    Parameters
    ----------
    axes: numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return np.array([normalize(a) for a in axes], dtype=axes.dtype)


def Gram(axes: ndarray) -> ndarray:
    """
    Returns the Gram matrix of a frame.

    Parameters
    ----------
    axes: numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return axes @ transpose_axes(axes)


def dual_frame(axes: ndarray) -> ndarray:
    """
    Returns the dual frame of the input.

    Parameters
    ----------
    axes: numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return transpose_axes(np.linalg.inv(axes))


def random_pos_semidef_matrix(N) -> ndarray:
    """
    Returns a random positive semidefinite matrix of shape (N, N).

    Example
    -------
    >>> from sigmaepsilon.math.linalg import random_pos_semidef_matrix, is_pos_semidef
    >>> arr = random_pos_semidef_matrix(2)
    >>> is_pos_semidef(arr)
    True
    """
    A = np.random.rand(N, N)
    return A.T @ A


def random_posdef_matrix(N, alpha: float = 1e-12) -> ndarray:
    """
    Returns a random positive definite matrix of shape (N, N).

    All eigenvalues of this matrix are >= alpha.

    Example
    -------
    >>> from sigmaepsilon.math.linalg import random_posdef_matrix, is_pos_def
    >>> arr = random_posdef_matrix(2)
    >>> is_pos_def(arr)
    True
    """
    A = np.random.rand(N, N)
    return A @ A.T + alpha * np.eye(N)


def inv_sym_3x3(m: Matrix, as_adj_det=False) -> Matrix:
    P11, P12, P13, P21, P22, P23, P31, P32, P33 = symbols(
        "P_{11} P_{12} P_{13} P_{21} P_{22} P_{23} P_{31} \
                P_{32} P_{33}",
        real=True,
    )
    Pij = [[P11, P12, P13], [P21, P22, P23], [P31, P32, P33]]
    P = sy.Matrix(Pij)
    detP = P.det()
    adjP = P.adjugate()
    invP = adjP / detP
    subs = {s: r for s, r in zip(sy.flatten(P), sy.flatten(m))}
    if as_adj_det:
        return detP.subs(subs), adjP.subs(subs)
    else:
        return invP.subs(subs)


@njit(nogil=True, parallel=True, cache=__cache)
def vpath(p1: ndarray, p2: ndarray, n: int) -> ndarray:
    nD = len(p1)
    dist = p2 - p1
    length = np.linalg.norm(dist)
    s = np.linspace(0, length, n)
    res = np.zeros((n, nD), dtype=p1.dtype)
    d = dist / length
    for i in prange(n):
        res[i] = p1 + s[i] * d
    return res


@njit(nogil=True, cache=__cache)
def linsolve(A, b) -> ndarray:
    return np.linalg.solve(A, b)


@njit(nogil=True, cache=__cache)
def inv(A: ndarray) -> ndarray:
    return np.linalg.inv(A)


@njit(nogil=True, cache=__cache)
def matmul(A: ndarray, B: ndarray) -> ndarray:
    return A @ B


@njit(nogil=True, cache=__cache)
def ATB(A: ndarray, B: ndarray) -> ndarray:
    return A.T @ B


@njit(nogil=True, cache=__cache)
def matmulw(A: ndarray, B: ndarray, w: float = 1.0) -> ndarray:
    return w * (A @ B)


@njit(nogil=True, cache=__cache)
def ATBA(A: ndarray, B: ndarray) -> ndarray:
    return A.T @ B @ A


@njit(nogil=True, cache=__cache)
def ATBAw(A: ndarray, B: ndarray, w: float = 1.0) -> ndarray:
    return w * (A.T @ B @ A)


@guv(["(f8[:, :], f8)"], "(n, n) -> ()", nopython=True, cache=__cache)
def det3x3(A, res):
    res = (
        A[0, 0] * A[1, 1] * A[2, 2]
        - A[0, 0] * A[1, 2] * A[2, 1]
        - A[0, 1] * A[1, 0] * A[2, 2]
        + A[0, 1] * A[1, 2] * A[2, 0]
        + A[0, 2] * A[1, 0] * A[2, 1]
        - A[0, 2] * A[1, 1] * A[2, 0]
    )


@guv(["(f8[:, :], f8)"], "(n, n) -> ()", nopython=True, cache=__cache)
def det2x2(A, res):
    res = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]


@njit(nogil=True, cache=__cache)
def inv2x2(A) -> ndarray:
    res = np.zeros_like(A)
    d = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res[0, 0] = A[1, 1] / d
    res[1, 1] = A[0, 0] / d
    res[0, 1] = -A[0, 1] / d
    res[1, 0] = -A[1, 0] / d
    return res


@guv(["(f8[:, :], f8[:, :])"], "(n, n) -> (n, n)", nopython=True, cache=__cache)
def inv2x2u(A, res):
    d = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res[0, 0] = A[1, 1] / d
    res[1, 1] = A[0, 0] / d
    res[0, 1] = -A[0, 1] / d
    res[1, 0] = -A[1, 0] / d


@guv(["(f8[:, :], f8[:, :])"], "(n, n) -> (n, n)", nopython=True, cache=__cache)
def adj3x3(A, res):
    res[0, 0] = A[1, 1] * A[2, 2] - A[1, 2] * A[2, 1]
    res[0, 1] = -A[0, 1] * A[2, 2] + A[0, 2] * A[2, 1]
    res[0, 2] = A[0, 1] * A[1, 2] - A[0, 2] * A[1, 1]
    res[1, 0] = -A[1, 0] * A[2, 2] + A[1, 2] * A[2, 0]
    res[1, 1] = A[0, 0] * A[2, 2] - A[0, 2] * A[2, 0]
    res[1, 2] = -A[0, 0] * A[1, 2] + A[0, 2] * A[1, 0]
    res[2, 0] = A[1, 0] * A[2, 1] - A[1, 1] * A[2, 0]
    res[2, 1] = -A[0, 0] * A[2, 1] + A[0, 1] * A[2, 0]
    res[2, 2] = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]


@guv(["(f8[:, :], f8[:, :])"], "(n, n) -> (n, n)", nopython=True, cache=__cache)
def inv3x3u(A, res):
    d = (
        A[0, 0] * A[1, 1] * A[2, 2]
        - A[0, 0] * A[1, 2] * A[2, 1]
        - A[0, 1] * A[1, 0] * A[2, 2]
        + A[0, 1] * A[1, 2] * A[2, 0]
        + A[0, 2] * A[1, 0] * A[2, 1]
        - A[0, 2] * A[1, 1] * A[2, 0]
    )
    res[0, 0] = A[1, 1] * A[2, 2] / d - A[1, 2] * A[2, 1] / d
    res[0, 1] = -A[0, 1] * A[2, 2] / d + A[0, 2] * A[2, 1] / d
    res[0, 2] = A[0, 1] * A[1, 2] / d - A[0, 2] * A[1, 1] / d
    res[1, 0] = -A[1, 0] * A[2, 2] / d + A[1, 2] * A[2, 0] / d
    res[1, 1] = A[0, 0] * A[2, 2] / d - A[0, 2] * A[2, 0] / d
    res[1, 2] = -A[0, 0] * A[1, 2] / d + A[0, 2] * A[1, 0] / d
    res[2, 0] = A[1, 0] * A[2, 1] / d - A[1, 1] * A[2, 0] / d
    res[2, 1] = -A[0, 0] * A[2, 1] / d + A[0, 1] * A[2, 0] / d
    res[2, 2] = A[0, 0] * A[1, 1] / d - A[0, 1] * A[1, 0] / d


@njit(nogil=True, cache=__cache)
def inv3x3(A):
    res = np.zeros_like(A)
    det = (
        A[0, 0] * A[1, 1] * A[2, 2]
        - A[0, 0] * A[1, 2] * A[2, 1]
        - A[0, 1] * A[1, 0] * A[2, 2]
        + A[0, 1] * A[1, 2] * A[2, 0]
        + A[0, 2] * A[1, 0] * A[2, 1]
        - A[0, 2] * A[1, 1] * A[2, 0]
    )
    res[0, 0] = A[1, 1] * A[2, 2] - A[1, 2] * A[2, 1]
    res[0, 1] = -A[0, 1] * A[2, 2] + A[0, 2] * A[2, 1]
    res[0, 2] = A[0, 1] * A[1, 2] - A[0, 2] * A[1, 1]
    res[1, 0] = -A[1, 0] * A[2, 2] + A[1, 2] * A[2, 0]
    res[1, 1] = A[0, 0] * A[2, 2] - A[0, 2] * A[2, 0]
    res[1, 2] = -A[0, 0] * A[1, 2] + A[0, 2] * A[1, 0]
    res[2, 0] = A[1, 0] * A[2, 1] - A[1, 1] * A[2, 0]
    res[2, 1] = -A[0, 0] * A[2, 1] + A[0, 1] * A[2, 0]
    res[2, 2] = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res /= det
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def inv3x3_bulk(A) -> ndarray:
    res = np.zeros_like(A)
    for i in prange(A.shape[0]):
        det = (
            A[i, 0, 0] * A[i, 1, 1] * A[i, 2, 2]
            - A[i, 0, 0] * A[i, 1, 2] * A[i, 2, 1]
            - A[i, 0, 1] * A[i, 1, 0] * A[i, 2, 2]
            + A[i, 0, 1] * A[i, 1, 2] * A[i, 2, 0]
            + A[i, 0, 2] * A[i, 1, 0] * A[i, 2, 1]
            - A[i, 0, 2] * A[i, 1, 1] * A[i, 2, 0]
        )
        res[i, 0, 0] = A[i, 1, 1] * A[i, 2, 2] - A[i, 1, 2] * A[i, 2, 1]
        res[i, 0, 1] = -A[i, 0, 1] * A[i, 2, 2] + A[i, 0, 2] * A[i, 2, 1]
        res[i, 0, 2] = A[i, 0, 1] * A[i, 1, 2] - A[i, 0, 2] * A[i, 1, 1]
        res[i, 1, 0] = -A[i, 1, 0] * A[i, 2, 2] + A[i, 1, 2] * A[i, 2, 0]
        res[i, 1, 1] = A[i, 0, 0] * A[i, 2, 2] - A[i, 0, 2] * A[i, 2, 0]
        res[i, 1, 2] = -A[i, 0, 0] * A[i, 1, 2] + A[i, 0, 2] * A[i, 1, 0]
        res[i, 2, 0] = A[i, 1, 0] * A[i, 2, 1] - A[i, 1, 1] * A[i, 2, 0]
        res[i, 2, 1] = -A[i, 0, 0] * A[i, 2, 1] + A[i, 0, 1] * A[i, 2, 0]
        res[i, 2, 2] = A[i, 0, 0] * A[i, 1, 1] - A[i, 0, 1] * A[i, 1, 0]
        res[i] /= det
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def inv3x3_bulk2(A) -> ndarray:
    res = np.zeros_like(A)
    for i in prange(A.shape[0]):
        res[i] = inv3x3(A[i])
    return res


@njit(nogil=True, cache=__cache)
def normalize(A) -> ndarray:
    return A / np.linalg.norm(A)


@njit(nogil=True, parallel=True, cache=__cache)
def normalize2d(A) -> ndarray:
    res = np.zeros_like(A)
    for i in prange(A.shape[0]):
        res[i] = normalize(A[i])
    return res


@njit(nogil=True, cache=__cache)
def norm(A) -> float:
    return np.linalg.norm(A)


@njit(nogil=True, parallel=True, cache=__cache)
def norm2d(A) -> ndarray:
    res = np.zeros(A.shape[0])
    for i in prange(A.shape[0]):
        res[i] = norm(A[i, :])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _linspace(p0: ndarray, p1: ndarray, N):
    s = p1 - p0
    L = np.linalg.norm(s)
    n = s / L
    djac = L / (N - 1)
    step = n * djac
    res = np.zeros((N, p0.shape[0]))
    res[0] = p0
    for i in prange(1, N - 1):
        res[i] = p0 + i * step
    res[-1] = p1
    return res


def linspace(start, stop, N) -> ndarray:
    if isinstance(start, ndarray):
        return _linspace(start, stop, N)
    else:
        return np.linspace(start, stop, N)


@njit(nogil=True, parallel=True, cache=__cache)
def linspace1d(start, stop, N) -> ndarray:
    res = np.zeros(N)
    di = (stop - start) / (N - 1)
    for i in prange(N):
        res[i] = start + i * di
    return res


def generalized_left_inverse(matrix: ndarray) -> ndarray:
    """Returns the generalized left inverse

    .. math::
        :nowrap:

        \\begin{equation}
            \left( \mathbf{A}^{T} \mathbf{A} \\right)^{-1} \mathbf{A}^{T}
        \\end{equation}

    """
    return np.linalg.inv(matrix.T @ matrix) @ matrix.T


def generalized_right_inverse(matrix: ndarray) -> ndarray:
    """Returns the generalized right inverse

    .. math::
        :nowrap:

        \\begin{equation}
            \mathbf{A}^{T} \left( \mathbf{A} \mathbf{A}^{T} \\right)^{-1}
        \\end{equation}

    """
    return matrix.T @ np.linalg.inv(matrix @ matrix.T)


def generalized_inverse(matrix: ndarray) -> ndarray:
    """
    Returns the generalized inverse of the input matrix, in any of the following
    cases:
    1) The matrix is square and has full rank. In this case the returned matrix
    is the usual inverse.
    2) The matrix has more columns than rows and has full row rank. In this case
    the generalized right inverse is returned.
    3) The matrix has more rows than columns and has full column rank. In this case
    the generalized left inverse is returned.
    """
    if not len(matrix.shape) == 2:
        raise LinalgOperationInputError("The input must be a matrix")

    num_rows, num_columns = matrix.shape
    rank = np.linalg.matrix_rank(matrix)
    if (num_rows == num_columns) and rank == num_columns == num_rows:
        return np.linalg.inv(matrix)
    elif (num_rows > num_columns) and rank == num_columns:
        return generalized_left_inverse(matrix)
    elif (num_rows < num_columns) and rank == num_rows:
        return generalized_right_inverse(matrix)
    else:
        raise LinalgError("The matrix has no inverse")



from abc import abstractmethod
from typing import Iterable, Callable, Tuple, List, Iterable, Optional, Union, Generator
import numpy as np
from numpy import ndarray
from pydantic import BaseModel, Field
from joblib import Parallel, delayed

__all__ = ["GeneticAlgorithm", "Genom"]


def even(n):
    return n % 2 == 0


def odd(n):
    return not even(n)


class Genom(BaseModel):
    """
    A data class for members of a population.
    """

    phenotype: List[float] = Field(default_factory=list)
    genotype: List[int] = Field(default_factory=list)
    fittness: float
    age: int = Field(default=0)
    index: int = Field(default=-1)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Genom):
            return False
        return np.all(self.genotype == other.genotype)

    def __hash__(self):
        arr_string = "".join(str(i) for i in self.genotype)
        return hash(arr_string)

    def __gt__(self, other):
        if not isinstance(other, Genom):
            raise TypeError(
                f"This operation is not supported between instances of {type(other)} and {type(self)}."
            )
        return np.all(self.fittness > other.fittness)

    def __lt__(self, other):
        if not isinstance(other, Genom):
            raise TypeError(
                f"This operation is not supported between instances of {type(other)} and {type(self)}."
            )
        return np.all(self.fittness < other.fittness)

    def __gte__(self, other):
        if not isinstance(other, Genom):
            raise TypeError(
                f"This operation is not supported between instances of {type(other)} and {type(self)}."
            )
        return np.all(self.fittness >= other.fittness)

    def __lte__(self, other):
        if not isinstance(other, Genom):
            raise TypeError(
                f"This operation is not supported between instances of {type(other)} and {type(self)}."
            )
        return np.all(self.fittness <= other.fittness)


class GeneticAlgorithm:
    """
    Base class for Genetic Algorithms (GA). Use this as a base
    class to your custom implementation of a GA.

    The class has 4 abstract methods wich upon being implemented, yield
    a working genetic algorithm. These are :func:`populate`, :func:`decode`,
    :func:`crossover`, :func:`mutate` and :func:`select`. It is also possible to
    use a custom stopping criteria by implementing :func:`stopping_criteria`.
    See :class:`~sigmaepsilon.math.optimize.ga.BinaryGeneticAlgorithm` for an example.

    Parameters
    ----------
    fnc: Callable
        The function to evaluate. It is assumed, that the function expects and N
        number of scalar arguments as a 1d iterable.
    ranges: Iterable
        Ranges for each scalar argument to the objective function.
    length: int, Optional
        Chromosome length. The higher the value, the more precision. Default is 5.
    p_c: float, Optional
        Probability of crossover. Default is 1.
    p_m: float, Optional
        Probability of mutation. Default is 0.2.
    nPop: int, Optional
        The size of the population. Default is 100.
    maxiter: int, Optional
        The maximum number of iterations. Default is 200.
    miniter: int, Optional
        The minimum number of iterations. Default is 100.
    elitism: float or int, Optional
        Default is 1
    ftol: float, Optional
        Torelance for floating point operations. Default is 1e-12.
    maxage: int, Optional
        The age is the number of generations a candidate spends at the top
        (being the best candidate). Setting an upper limit to this value is a kind
        of stopping criterion. Default is 5.
    num_proc: int, Optional
        The number of workers used to evaluate the population. Use a value of `-1` to use
        all cores. Default is 1.

        .. versionadded:: 1.1.0

    Note
    ----
    Be cautious what you use a genetic algorithm for. Like all metahauristic methods, a
    genetic algorithm can be wery demanding on the computational side. If the objective
    unction takes a lot of time to evaluate, it is probably not a good idea to use a heuristic
    approach, unless you have a dedicated solver that is able to run efficiently for a large
    number of problems.

    See also
    --------
    :class:`~sigmaepsilon.math.optimize.ga.BinaryGeneticAlgorithm`
    """

    def __init__(
        self,
        fnc: Callable,
        ranges: Iterable,
        *,
        length: Optional[int] = 5,
        p_c: Optional[float] = 1,
        p_m: Optional[float] = 0.2,
        nPop: Optional[int] = 100,
        maxiter: Optional[int] = 200,
        miniter: Optional[int] = 0,
        elitism: Optional[int] = 1,
        maxage: Optional[int] = 5,
        num_proc: Optional[int] = 1,
    ):
        super().__init__()
        self.fnc = fnc
        self.ranges = np.array(ranges)
        self.dim = getattr(fnc, "dimension", self.ranges.shape[0])
        self.length = length
        self.p_c = p_c
        self.p_m = p_m

        if odd(nPop):
            nPop += 1
        if odd(int(nPop / 2)):
            nPop += 2
        assert nPop % 4 == 0
        assert nPop >= 4

        self.nPop = nPop
        self._genotypes = None
        self._pnenotypes = None
        self._fittness = None
        self._champion: Genom = None
        self.reset()
        self._set_solution_params(
            maxiter=maxiter,
            miniter=miniter,
            elitism=elitism,
            maxage=maxage,
            num_proc=num_proc,
        )

        if self.miniter > self.maxiter:
            raise ValueError("'maxiter' must be greater than 'miniter'")

    @property
    def champion(self) -> Genom:
        """
        Returnes the genotypes of the population.
        """
        return self._champion

    @property
    def genotypes(self) -> Iterable:
        """
        Returnes the genotypes of the population.
        """
        return self._genotypes

    @genotypes.setter
    def genotypes(self, value: Iterable):
        """
        Sets the genotypes of the population.
        """
        self._genotypes = value
        self._phenotypes = None
        self._fittness = None

    @property
    def phenotypes(self) -> Iterable:
        """
        Returnes the phenotypes of the population.
        """
        if self._phenotypes is None:
            genotypes = self.genotypes
            if genotypes is not None:
                self._phenotypes = self.decode(genotypes)
        return self._phenotypes

    @property
    def fittness(self) -> ndarray:
        """
        Returns the actual fittness values of the population, or the fittness
        of the population described by the argument `phenotypes`.
        """
        if self._fittness is None:
            self._fittness = self.evaluate(self.phenotypes)
        return self._fittness

    def reset(self) -> "GeneticAlgorithm":
        """
        Resets the solver and returns the object. Only use it if you want to have
        a completely clean sheet. Also, the function is called for every object at
        instantiation.
        """
        self._evolver = self.evolver()
        self._evolver.send(None)
        self._champion = None
        return self

    def _set_solution_params(
        self,
        *,
        maxiter: Optional[Union[int, None]] = None,
        miniter: Optional[Union[int, None]] = None,
        elitism: Optional[Union[int, None]] = None,
        maxage: Optional[Union[int, None]] = None,
        num_proc: Optional[Union[int, None]] = None,
    ) -> "GeneticAlgorithm":
        if maxiter is not None:
            self.maxiter = maxiter
        if miniter is not None:
            self.miniter = miniter
        if elitism is not None:
            self.elitism = elitism
        if maxage is not None:
            self.maxage = maxage
        if num_proc is not None:
            self.num_proc = num_proc
        return self

    def evolver(self) -> Iterable:
        """
        Returns a generator that can be used to manually control evolutions.
        """
        self.genotypes = self.populate()
        _ = yield
        yield self.genotypes
        while True:
            self.genotypes = self.populate(
                self.select(self._genotypes, self.phenotypes)
            )
            yield self._genotypes

    def evolve(self, cycles: Optional[int] = 1) -> Iterable:
        """
        Performs a certain number of cycles of evolution and returns the
        genotypes.
        """
        for _ in range(cycles):
            next(self._evolver)
        return self.genotypes

    def solve(self, recycle: Optional[bool] = False, **kwargs) -> Genom:
        """
        Solves the problem and returns the best phenotype.

        Parameters
        ----------
        recycle: bool, Optional
            If True, the leftover resources of the previous calculation are the starting
            point of the new solution.

        Returns
        -------
        :class:`~sigmaepsilon.math.optimize.ga.Genom`
            The best candidate.
        """
        self.reset() if not recycle else None
        self._set_solution_params(**kwargs)

        nIter, finished = 0, False

        while (not finished and nIter < self.maxiter) or (nIter < self.miniter):
            self.evolve()
            candidate: Genom = self.best_candidate()
            self.celebrate(candidate)
            finished = self.stopping_criteria()
            nIter += 1

        self.nIter = nIter
        return self.champion

    def evaluate(self, phenotypes: Optional[Union[Iterable, None]] = None) -> ndarray:
        """
        Evaluates the objective for a list of phenotypes.

        If the phenotypes are not explicitly specified, the population at hand
        is evaluated.

        Parameters
        ----------
        phenotypes: Iterable, Optional
            The phenotypes the objective function is to be evaluated for.
            Default is None.
        """
        phenotypes = self.phenotypes if phenotypes is None else phenotypes
        if self.num_proc == 1:
            return np.array([self.fnc(x) for x in phenotypes], dtype=float)
        else:
            results = Parallel(n_jobs=self.num_proc)(
                delayed(self.fnc)(x) for x in phenotypes
            )
            return np.array(results, dtype=float)

    def best_phenotype(self) -> ndarray:
        """
        Returns the best phenotype.

        Parameters
        ----------
        lastknown: bool, Optional
            If True, the last evaluation is used. If False, the phenotypes
            are evaluated before selecting the best. In this case the results
            are not stored. Default is True.
        """
        return self.best_candidate().phenotype

    def best_candidate(self) -> Genom:
        """
        Returns data about the best candidate in the population like index,
        phenotype, genotype and fittness value.

        Parameters
        ----------
        lastknown: bool, Optional
            If True, the last evaluation is used. If False, the phenotypes
            are evaluated before selecting the best. In this case the results
            are not stored. Default is True.
        """
        fittness = self.fittness
        index = np.argmin(fittness)
        return Genom(
            phenotype=self.phenotypes[index],
            genotype=self.genotypes[index],
            fittness=fittness[index],
            index=index,
        )

    def celebrate(self, genom: Genom) -> None:
        """
        Celebration of the winner. Curretly this means that the beast candidate is added
        to a history to keep track of the improvements across evolutions.
        """
        if self.champion is None:
            self._champion = genom
            self._champion.age = 0
        else:
            if genom > self.champion:
                self._champion = genom
                self._champion.age = 0
        self._champion.age += 1

    def divide(self, fittness: Optional[Union[ndarray, None]] = None) -> Tuple[List]:
        """
        Divides population to elit and others, and returns the corresponding
        index arrays.

        Parameters
        ----------
        fittness: numpy.ndarray, Optional
            Fittness values. If not provided, values from the latest
            evaluation are used. Default is None.

        Returns
        -------
        list
            Indices of the members of the elite.
        list
            Indices of the members of the others.
        """
        fittness = self.fittness if fittness is None else fittness
        assert fittness is not None, "No available fittness data detected."
        if self.elitism < 1:
            argsort = np.argsort(fittness)
            elit = argsort[: int(self.nPop * self.elitism)]
            others = argsort[int(self.nPop * self.elitism) :]
        elif self.elitism > 1:
            argsort = np.argsort(fittness)
            elit = argsort[: self.elitism]
            others = argsort[self.elitism :]
        else:
            elit = []
            others = list(range(self.nPop))
        return list(elit), others

    @classmethod
    def random_parents_generator(
        cls, genotypes: Optional[Union[ndarray, None]] = None
    ) -> Generator:
        """
        Yields random pairs from a list of genotypes.

        The implemantation assumes that the length of the input array
        is a multiple of 2.

        Parameters
        ----------
        genotypes: numpy.ndarray
            Genotypes of the parents as a 2d integer array.

        Yields
        ------
        numpy.ndarray
            The first parent.
        numpy.ndarray
            The second parent.
        """
        n = len(genotypes)
        assert n % 2 == 0, "'n' must be a multiple of 2"
        pool = np.full(n, True)
        nPool = n
        while nPool > 2:
            where = np.argwhere(pool == True).flatten()
            nPool = len(where)
            pair = np.random.choice(where, 2, replace=False)
            parent1 = genotypes[pair[0]]
            parent2 = genotypes[pair[1]]
            pool[pair] = False
            yield parent1, parent2

    @abstractmethod
    def stopping_criteria(self) -> bool:
        """
        Implements a simple stopping criteria that evaluates to `True` if the
        current chanpion is thought ti bee the best solution and no further progress
        can be made, or at lest with a bad rate.

        The default implementation considers a champion as the winner, if it is the champion
        for for at least 5 times in a row. This can be dontrolled with the `maxage` parameter
        when instantiating an instance.
        """
        return self.champion.age > self.maxage

    @abstractmethod
    def encode(self, phenotypes: Optional[Union[ndarray, None]] = None) -> ndarray:
        """
        Turns phenotypes into genotypes.
        """
        return phenotypes

    @abstractmethod
    def decode(self, genotypes: Optional[Union[ndarray, None]] = None) -> ndarray:
        """
        Turns genotypes into phenotypes.
        """
        return genotypes

    @abstractmethod
    def populate(self, genotypes: Optional[Union[ndarray, None]] = None):
        """
        Ought to produce a pool of phenotypes.
        """
        ...

    @abstractmethod
    def crossover(
        self,
        parent1: Optional[Union[ndarray, None]] = None,
        parent2: Optional[Union[ndarray, None]] = None,
    ) -> Tuple[ndarray]:
        """
        Takes in two parents, returns two offspring. You'd probably want to use it inside
        the populator.
        """
        ...

    @abstractmethod
    def mutate(self, child: Optional[Union[ndarray, None]] = None) -> ndarray:
        """
        Takes a child in, returns a mutant.
        """
        ...

    @abstractmethod
    def select(
        self,
        genotypes: Optional[Union[ndarray, None]] = None,
        phenotypes: Optional[Union[ndarray, None]] = None,
    ):
        """
        Ought to implement dome kind of selection mechanism, eg. a roulette wheel,
        tournament or other.
        """
        ...



import sympy as sy

from .testfunction import TestMinFunction2D
from .metafunction import symbolize, substitute


def Rosenbrock(a=1, b=100) -> TestMinFunction2D:
    """
    Implements the Rosenbrock function, aka. Banana function.

    :math:`f(x, y) = (a - x)^2 + b (y - x^2)^2`

    a,b are constants, tipically a is set to 1 and  b is set to 100.

    One global minimum : f(1, 1) = 0.

    """
    str_expr = "(a-x)**2 + b*(y-x**2)**2"
    expr = substitute(sy.sympify(str_expr), [a, b], ["a", "b"])
    return TestMinFunction2D(**symbolize(expr=expr), optimums=[(1.0, 1.0)])


def Himmelblau() -> TestMinFunction2D:
    """
    Creates the Himmelblau's function object.

    :math:`f(x,y) = (x^2 + y - 11)^2 + (x + y^2 - 7)^2`

    Four identical local minima with values 0.0 at
    (3.0,2.0), (-2.805,3.131), (-3.779,-3.283), (3.584,-1.848).

    """
    str_expr = "(x**2 + y - 11)**2 + (x + y**2 - 7)**2"
    return TestMinFunction2D(
        **symbolize(str_expr=str_expr),
        optimums=[
            (3.0, 2.0),
            (-2.805118, 3.131312),
            (-3.779310, -3.283186),
            (3.584428, -1.848126),
        ]
    )


def GoldsteinPrice() -> TestMinFunction2D:
    """
    Creates the Goldstein-Price function object.

    .. math::
        :nowrap:

        \\begin{equation}
            f(x, y) = (1 + (x + y + 1)^2  (19-14x + 3x^2 - 14y + 6xy + 3y^2))
            (30 + (2x - 3y)^2  (18 - 32x + 12x^2 + 48y - 36xy + 27y^2))
        \\end{equation}


    One global minimum : f(0,-1) = 3.

    """
    str_expr = (
        "(1+(x+y+1)**2 * (19-14*x+3*x**2-14*y+6*x*y+3*y**2))"
        "*(30+(2*x-3*y)**2 * (18-32*x+12*x**2+48*y-36*x*y+27*y**2))"
    )
    return TestMinFunction2D(**symbolize(str_expr=str_expr), optimums=[(0.0, -1.0)])


def Beale() -> TestMinFunction2D:
    """
    Creates the Beale function object.

    .. math::
        :nowrap:

        \\begin{equation}
            f(x, y) = (1.5 - x + x y)^2 + (2.25 - x + x y^2)^2
            + (2.625 - x + x y^3)^2
        \\end{equation}


    One global minimum : f(3,0.5) = 0.

    """
    str_expr = (
        "(1.5 - x + x*y)**2 + (2.25 - x + x*y**2)**2 " "+ (2.625 - x + x*y**3)**2"
    )
    return TestMinFunction2D(**symbolize(str_expr=str_expr), optimums=[(3.0, 0.5)])


def Matyas() -> TestMinFunction2D:
    """
    Creates the Matyas function object.

    :math:`f(x, y) = 0.26 (x^2 + y^2) - 0.48 x y`

    One global minimum : f(0., 0.) = 0.

    """
    str_expr = "0.26*(x**2 + y**2) - 0.48*x*y"
    return TestMinFunction2D(**symbolize(str_expr=str_expr), optimums=[(0.0, 0.0)])


if __name__ == "__main__":
    """
    check:
        - Himmelblau with initial = [-1.,-1.]. Problem with the Hessian
          at [-0.270845, -0.923039].

    """



from sigmaepsilon.core.exceptions import SigmaEpsilonException


class TensorShapeMismatchError(SigmaEpsilonException):
    """Mismatch in the shape of the inputs."""


class LinalgOperationInputError(SigmaEpsilonException):
    """Invalid input for this operation."""


class LinalgMissingInputError(SigmaEpsilonException):
    """Invalid input for this operation."""


class LinalgInvalidTensorOperationError(SigmaEpsilonException):
    """
    Tensors don't support this operation. Try to call this using
    the arrays of the tensorial inputs.
    """


class LinalgError(SigmaEpsilonException):
    """General linear algebra error"""



# -*- coding: utf-8 -*-
import numpy as np
from numba import jit

from ..sparse import CSR
from ...topology.graph import pseudo_peripheral_nodes


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def optimal_node_order(adj: CSR):
    itype = adj.indptr
    nN = len(adj.indptr) - 1  # number of nodes
    roots = pseudo_peripheral_nodes(adj)
    renum = np.full(nN, -1, dtype=itype)
    order = np.arange(0, nN, dtype=itype)
    minwidth = nN

    # 0  == virgin
    # -1 == eliminated
    # n  == age
    status = np.zeros(nN, dtype=itype)

    def iseliminated(status, i):
        return status[i] < 0

    def isactive(status, i):
        return status[i] > 0

    def isvirgin(status, i):
        return status[i] == 0

    def eliminate(status, i):
        status[i] = -1

    def activate(status, i):
        status[i] = 1

    def get_neighbours_of(adj, i):
        _, inds = adj.row(i)
        return inds

    def get_actives(status):
        return np.argwhere(status > 0).flatten()

    def get_cost(adj, status, i):
        cost = -1
        neighbours = get_neighbours_of(adj, i)
        for node in neighbours:
            if isvirgin(status, node):
                cost += 1
        return cost

    def get_costs(adj, status, actives):
        # measure cost of the elimination of active nodes
        nA = len(actives)
        costs = np.full(nA, -1, dtype=np.float64)
        for iA in range(nA):
            costs[iA] = get_cost(adj, status, actives[iA])
        return costs

    def get_cheapest(adj, status, actives):
        costs = get_costs(adj, status, actives)
        cheapest = np.argwhere(costs == np.min(costs)).flatten()
        # decide draw based on time spent in the front
        if len(cheapest) > 1:
            ages = status[actives[cheapest]]
            aind = cheapest[np.argmax(ages)]
        else:
            aind = cheapest[0]
        return actives[aind]

    nN -= 1  # for direct comparison with running variable
    nR = len(roots)
    # the root of the level structure is the last item in roots
    for iR in range(nR - 1, -1, -1):
        # initialize
        status[:] = 0
        renum[:] = -1
        nextindex = 0
        width = 0

        # register starting node, mark all neighbours as active
        root = roots[iR]
        renum[root] = nextindex
        eliminate(status, root)
        neighbours = get_neighbours_of(adj, root)
        for node in neighbours:
            if isvirgin(status, node):
                activate(status, node)

        while nN > nextindex:  # eliminate until there is nothing left
            # measure current solution
            actives = get_actives(status)
            width = max(width, len(actives))
            if width >= minwidth:
                break

            # increase age of active variables
            for node in actives:
                status[node] += 1

            # get variable cheapest to eliminate
            cheapest = get_cheapest(adj, status, actives)

            # eliminate variable and read adjacent variables
            nextindex += 1
            renum[cheapest] = nextindex
            eliminate(status, cheapest)
            neighbours = get_neighbours_of(adj, cheapest)
            for node in neighbours:
                if isvirgin(status, node):
                    activate(status, node)

        # register maximum occuring width
        if np.min(renum) == 0:
            if width < minwidth:
                order[:] = renum
                minwidth = width
    return order


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def optimal_node_path(adj: CSR = None):
    return order_to_path(optimal_node_order(adj))


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def optimal_elimination_order(adj: CSR, topo: CSR):
    # renumber nodes
    norder = optimal_node_order(adj)
    for i in range(len(topo.data)):
        topo.data[i] = norder[topo.data[i]]

    # reorder elements
    nN = len(adj.indptr) - 1  # number of nodes
    nE = len(topo.indptr) - 1  # number of elements
    eorder = np.full(nE, -1, dtype=np.int64)
    cE = 0
    for iN in range(nN):
        for iE in range(nE):
            if eorder[iE] > -1:
                continue
            else:
                nodes, _ = topo.row(iE)
                for node in nodes:
                    if node == iN:
                        eorder[iE] = cE
                        cE += 1
                        break

    # undo renumbering of nodes
    npath = order_to_path(norder)
    for i in range(len(topo.data)):
        topo.data[i] = npath[topo.data[i]]
    return eorder


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def optimal_elimination_path(adj: CSR, topo: CSR):
    return order_to_path(optimal_elimination_order(adj, topo))


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def order_to_path(order: np.ndarray):
    path = np.zeros(order.shape, dtype=order.dtype)
    for i in range(len(path)):
        path[order[i]] = i
    return path


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def path_to_order(path: np.ndarray):
    order = np.zeros(path.shape, dtype=path.dtype)
    for i in range(len(order)):
        order[path[i]] = i
    return order


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def reverse_mapping(iarray: np.ndarray):
    """
    Returns the inverse of a resequencing path represented by the
    complete integer array iarray. Here, completeness means that
    the array iarray must be a shuffled copy of np.arange(len(iarray)).
    """
    return order_to_path(iarray)


if __name__ == "__main__":
    pass
    """
    from polydata.mechanics.fem import Uniform3D
    from dewloosh.math.linalg.frontal.frutils import max_frontwidth_bulk
    from dewloosh.math.linalg.frontal.topo import signed_topo_bulk
    from dewloosh.math.linalg.sparse import CSR

    #####################
    # Example 1
    # basic operations
    #####################

    # geometry
    Lx = 100.0
    Ly = 100.0
    Lz = 100

    # mesh control
    nx = 21
    ny = 21
    nz = 21

    # assembly
    assembly = Uniform3D((Lx, Ly, Lz), (nx, ny, nz))
    points = assembly.points()
    cells = assembly.cells()

    # connectivity and topology
    topo = assembly.element_node_numbering()
    adj_csr = assembly.adjacency_matrix(sparse=True)
    topo_csr = assembly.connectivity_matrix(sparse=True)
    roots = pseudo_peripheral_nodes(adj_csr)
    norder = optimal_node_order(adj_csr)
    npath = order_to_path(norder)
    epath_0 = np.arange(len(cells))
    epath = optimal_elimination_path(adj_csr, topo_csr)
    eorder = optimal_elimination_order(adj_csr, topo_csr)
    topo_sig_0 = signed_topo_bulk(topo, epath_0)
    topo_sig = signed_topo_bulk(topo, epath)
    maxwidth = max_frontwidth_bulk(topo_sig, epath)
    maxwidth_0 = max_frontwidth_bulk(topo_sig_0, epath_0)
    print(maxwidth_0, ' --> ', maxwidth)

    #####################
    # Example 2
    # Hinton, Owen : Finite Element Programming, 1977, Academic Press,
    #     p182.
    # Statement : minimum frontwidth should be 12
    #####################

    topo_signed_1 = np.array([[-1, -2, -3, 4, 5, 6, 10, 11, 12],
                              [4, 5, 6, -13, -14, -15, 7, 8, 9],
                              [-4, -5, -6, -7, -8, -9, -10, -11, -12]])
    path_1 = np.arange(3, dtype=np.int64)
    maxwidth_topo_1 = max_frontwidth_bulk(topo_signed_1, path_1)
    """



__all__ = ["DegenerateProblemError", "NoSolutionError"]


class DegenerateProblemError(Exception):
    """
    The problem is degenerate.

    The objective could be decreased, but only on the expense
    of violating positivity of the standard variables.
    """

    ...


class NoSolutionError(Exception):
    """
    There is no solution to this problem.

    Step size could be indefinitely increased in a
    direction without violating feasibility.
    """

    ...



from collections import OrderedDict

import sympy as sy
from sympy import lambdify

from .metafunction import substitute


class VariableManager(object):
    def __init__(self, variables=None, vmap=None, **kwargs):
        try:
            variables = list(sy.symbols(variables, **kwargs))
        except Exception:
            variables = variables
        try:
            self.vmap = (
                vmap if vmap is not None else OrderedDict({v: v for v in variables})
            )
        except Exception:
            self.vmap = OrderedDict()
        self.variables = variables  # this may be unnecessary

    def substitute(self, vmap: dict = None, inverse=False, inplace=True):
        if not inverse:
            sval = list(vmap.values())
            svar = list(vmap.keys())
        else:
            sval = list(vmap.keys())
            svar = list(vmap.values())
        if inplace:
            for v, expr in self.vmap.items():
                self.vmap[v] = substitute(expr, sval, svar)
            return self
        else:
            vmap = OrderedDict()
            for v, expr in self.vmap.items():
                vmap[v] = substitute(expr, sval, svar)
            return vmap

    def lambdify(self, variables=None):
        assert variables is not None
        for v, expr in self.vmap.items():
            self.vmap[v] = lambdify([variables], expr, "numpy")

    def __call__(self, v):
        return self.vmap[v]

    def target(self):
        return list(self.vmap.keys())

    def source(self):
        s = set()
        for expr in self.vmap.values():
            s.update(expr.free_symbols)
        return list(s)

    def add_variables(self, variables, overwrite=True):
        if overwrite:
            self.vmap.update({v: v for v in variables})
        else:
            for v in variables:
                if v not in self.vmap:
                    self.vmap[v] = v